# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""The superclass of all handlers."""

import base64
import cgi
import datetime
import json
import logging
import os
import re
import sys
import traceback
import urllib.parse

from flask import redirect as flask_redirect
from flask import request
from flask import Response
from flask.views import MethodView
from google.cloud import ndb
import jinja2

from clusterfuzz._internal.base import utils
from clusterfuzz._internal.config import db_config
from clusterfuzz._internal.config import local_config
from clusterfuzz._internal.google_cloud_utils import storage
from clusterfuzz._internal.system import environment
from libs import auth
from libs import form
from libs import helpers

# Pattern from
# https://github.com/google/closure-library/blob/
# 3037e09cc471bfe99cb8f0ee22d9366583a20c28/closure/goog/html/safeurl.js
_SAFE_URL_PATTERN = re.compile(
    r'^(?:(?:https?|mailto|ftp):|[^:/?#]*(?:[/?#]|$))', flags=re.IGNORECASE)


def add_jinja2_filter(name, fn):
  _JINJA_ENVIRONMENT.filters[name] = fn


class JsonEncoder(json.JSONEncoder):
  """Json encoder."""
  _EPOCH = datetime.datetime.utcfromtimestamp(0)

  def default(self, obj):  # pylint: disable=arguments-differ,method-hidden
    if isinstance(obj, ndb.Model):
      dict_obj = obj.to_dict()
      dict_obj['id'] = obj.key.id()
      return dict_obj
    if isinstance(obj, datetime.datetime):
      return int((obj - self._EPOCH).total_seconds())
    if hasattr(obj, 'to_dict'):
      return obj.to_dict()
    if isinstance(obj, cgi.FieldStorage):
      return str(obj)
    if isinstance(obj, bytes):
      return obj.decode('utf-8')

    return json.JSONEncoder.default(self, obj)


def format_time(dt):
  """Format datetime object for display."""
  return '{t.day} {t:%b} {t:%y} {t:%X} PDT'.format(t=dt)


def splitlines(text):
  """Split text into lines."""
  return text.splitlines()


def split_br(text):
  return re.split(r'\s*<br */>\s*', text, flags=re.IGNORECASE)


def encode_json(value):
  """Dump base64-encoded JSON string (to avoid XSS)."""
  return base64.b64encode(json.dumps(
      value, cls=JsonEncoder).encode('utf-8')).decode('utf-8')


_JINJA_ENVIRONMENT = jinja2.Environment(
    loader=jinja2.FileSystemLoader(
        os.path.join(os.path.dirname(__file__), '..', 'templates')),
    extensions=['jinja2.ext.autoescape'],
    autoescape=True)
_MENU_ITEMS = []

add_jinja2_filter('json', encode_json)
add_jinja2_filter('format_time', format_time)
add_jinja2_filter('splitlines', splitlines)
add_jinja2_filter('split_br', split_br)
add_jinja2_filter('polymer_tag', lambda v: '{{%s}}' % v)


def add_menu(name, href):
  """Add menu item to the main navigation."""
  _MENU_ITEMS.append(_MenuItem(name, href))


def make_login_url(dest_url):
  """Make the switch account url."""
  return '/login?' + urllib.parse.urlencode({'dest': dest_url})


def make_logout_url(dest_url):
  """Make the switch account url."""
  return '/logout?' + urllib.parse.urlencode({
      'csrf_token': form.generate_csrf_token(),
      'dest': dest_url,
  })


def check_redirect_url(url):
  """Check redirect URL is safe."""
  if not _SAFE_URL_PATTERN.match(url):
    raise helpers.EarlyExitException('Invalid redirect.', 403)


class _MenuItem(object):
  """A menu item used for rendering an item in the main navigation."""

  def __init__(self, name, href):
    self.name = name
    self.href = href


class Handler(MethodView):
  """A superclass for all handlers. It contains many convenient methods."""

  def is_cron(self):
    """Return true if the request is from a cron job."""
    return bool(request.headers.get('X-Appengine-Cron'))

  def should_render_json(self):
    return (self.is_json or
            'application/json' in request.headers.get('accept', ''))

  def render_forbidden(self, message):
    """Write HTML response for 403."""
    login_url = make_login_url(dest_url=request.url)
    user_email = helpers.get_user_email()
    if not user_email:
      return self.redirect(login_url)

    contact_string = db_config.get_value('contact_string')
    template_values = {
        'message': message,
        'user_email': helpers.get_user_email(),
        'login_url': login_url,
        'switch_account_url': login_url,
        'logout_url': make_logout_url(dest_url=request.url),
        'contact_string': contact_string,
    }
    return self.render('error-403.html', template_values, 403)

  def _add_security_response_headers(self, response):
    """Add security-related headers to response."""
    response.headers['Strict-Transport-Security'] = (
        'max-age=2592000; includeSubdomains')
    response.headers['X-Content-Type-Options'] = 'nosniff'
    response.headers['X-Frame-Options'] = 'deny'
    return response

  def render(self, path, values=None, status=200):
    """Write HTML response."""
    if values is None:
      values = {}

    values['menu_items'] = _MENU_ITEMS
    values['is_oss_fuzz'] = utils.is_oss_fuzz()
    values['is_development'] = (
        environment.is_running_on_app_engine_development())
    values['is_logged_in'] = bool(helpers.get_user_email())

    # Only track analytics for non-admin users.
    values['ga_tracking_id'] = (
        local_config.GAEConfig().get('ga_tracking_id')
        if not auth.is_current_user_admin() else None)

    if values['is_logged_in']:
      values['switch_account_url'] = make_login_url(request.url)
      values['logout_url'] = make_logout_url(dest_url=request.url)

    template = _JINJA_ENVIRONMENT.get_template(path)

    response = Response()
    response = self._add_security_response_headers(response)
    response.headers['Content-Type'] = 'text/html'
    response.data = template.render(values)
    response.status_code = status
    return response

  # pylint: disable=unused-argument
  def before_render_json(self, values, status):
    """A hook for modifying values before render_json."""

  def render_json(self, values, status=200):
    """Write JSON response."""
    response = Response()
    response = self._add_security_response_headers(response)
    response.headers['Content-Type'] = 'application/json'
    self.before_render_json(values, status)
    response.data = json.dumps(values, cls=JsonEncoder)
    response.status_code = status
    return response

  def handle_exception(self, exception):
    """Catch exception and format it properly."""
    try:
      status = 500
      values = {
          'message': str(exception),
          'email': helpers.get_user_email(),
          'traceDump': traceback.format_exc(),
          'status': status,
          'type': exception.__class__.__name__
      }
      if isinstance(exception, helpers.EarlyExitException):
        status = exception.status
        values = exception.to_dict()

      # 4XX is not our fault. Therefore, we hide the trace dump and log on
      # the INFO level.
      if 400 <= status <= 499:
        logging.info(json.dumps(values, cls=JsonEncoder))
        del values['traceDump']
      else:  # Other error codes should be logged with the EXCEPTION level.
        logging.exception(exception)

      if self.should_render_json():
        return self.render_json(values, status)
      if status in (403, 401):
        return self.render_forbidden(str(exception))
      return self.render('error.html', values, status)
    except Exception:
      self.handle_exception_exception()

    return None

  def handle_exception_exception(self):
    """Catch exception in handle_exception and format it properly."""
    exception = sys.exc_info()[1]
    values = {'message': str(exception), 'traceDump': traceback.format_exc()}
    logging.exception(exception)
    if self.should_render_json():
      return self.render_json(values, 500)
    return self.render('error.html', values, 500)

  def redirect(self, url, **kwargs):
    """Check vaid url and redirect to it, if valid."""
    url = str(url)
    check_redirect_url(url)
    return flask_redirect(url, **kwargs)

  def dispatch_request(self, *args, **kwargs):
    """Dispatch a request and postprocess."""
    self.is_json = False
    try:
      return super(Handler, self).dispatch_request(*args, **kwargs)
    except Exception as exception:
      return self.handle_exception(exception)


class GcsUploadHandler(Handler):
  """A handler which uploads files to GCS."""

  def dispatch_request(self, *args, **kwargs):
    """Dispatch a request and postprocess."""
    self.upload = None
    return super().dispatch_request(*args, **kwargs)

  def get_upload(self):
    """Get uploads."""
    if self.upload:
      return self.upload

    upload_key = request.get('upload_key')
    if not upload_key:
      return None

    blob_info = storage.GcsBlobInfo.from_key(upload_key)
    if not blob_info:
      raise helpers.EarlyExitException('Failed to upload.', 500)

    self.upload = blob_info
    return self.upload



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""logger tests."""
import os
import unittest

from clusterfuzz._internal.platforms.android import logger
from clusterfuzz._internal.tests.test_libs import helpers as test_helpers

DATA_PATH = os.path.join(
    os.path.dirname(os.path.abspath(__file__)), 'logger_data')


class IsLineValidTest(unittest.TestCase):
  """Tests is_line_valid."""

  def test_beginning_of(self):
    """Tests beginning of line."""
    self.assertFalse(logger.is_line_valid('--------- beginning of system'))

  def test_debug(self):
    """Tests debug line."""
    self.assertFalse(
        logger.is_line_valid('D/ConnectivityService(  812): notifyType'))

  def test_verbose(self):
    """Tests verbose line."""
    self.assertFalse(
        logger.is_line_valid('V/chromium( 8572): [VERBOSE1:] Not implemented.'))

  def test_chromium_resource_load(self):
    """Tests verbose line."""
    self.assertTrue(
        logger.is_line_valid(
            'V/chromium( 8530): [VERBOSE1:network_delegate.cc(31)] '
            'NetworkDelegate::NotifyBeforeURLRequest: '
            'https://en.m.wikipedia.org'))

  def test_info(self):
    """Tests info line."""
    self.assertTrue(
        logger.is_line_valid(
            'I/chromium( 8530): [INFO:CONSOLE(166)] Hello world!'))


class FilterLogOutputTest(unittest.TestCase):
  """Tests filter_log_output."""

  def setUp(self):
    test_helpers.patch(self, ['clusterfuzz._internal.metrics.logs.log_error'])

  def _get_log_content(self, filename):
    return open(os.path.join(DATA_PATH, filename)).read()

  def test_sanitizer_and_check_stack(self):
    """Tests sanitizer check failure log output."""
    unfiltered_log_output = self._get_log_content(
        'check_failure_and_asan_log.txt')
    expected_filtered_log_output = self._get_log_content(
        'check_failure_and_asan_log_expected.txt')
    actual_filtered_log_output = logger.filter_log_output(unfiltered_log_output)

    self.assertEqual(actual_filtered_log_output, expected_filtered_log_output)
    self.assertEqual(0, self.mock.log_error.call_count)

  def test_process_with_type(self):
    """Tests log output where process has a type specifier."""
    self.assertEqual(
        '--------- EXT4-fs (loop25) (781):\n'
        'mounted filesystem without journal. Opts: (null)\n',
        logger.filter_log_output(
            'I/EXT4-fs (loop25)(  781): '
            'mounted filesystem without journal. Opts: (null)'))



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Revisions related helper functions."""

import ast
import base64
import bisect
import os
import re
import time
import urllib.parse

import six

from clusterfuzz._internal.base import memoize
from clusterfuzz._internal.base import utils
from clusterfuzz._internal.build_management import overrides
from clusterfuzz._internal.build_management import source_mapper
from clusterfuzz._internal.config import local_config
from clusterfuzz._internal.datastore import data_handler
from clusterfuzz._internal.google_cloud_utils import storage
from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.system import environment

CHROMIUM_GIT_ROOT_URL = 'https://chromium.googlesource.com'
CRREV_NUMBERING_URL = (
    'https://cr-rev.appspot.com/_ah/api/crrev/v1/get_numbering')
CLANK_URL = 'https://chrome-internal.googlesource.com/clank/internal/apps.git'
CLANK_REVISION_FILE_COMPONENT_REGEX = re.compile(
    r'.*["]([^"]+)["]\s*:\s*["]([^"]+)["]')
COMPONENT_NAMES_BLACKLIST = [
    'api', 'bin', 'data', 'dist', 'lib', 'pylib', 'source', 'src'
]
DISK_CACHE_SIZE = 1000
SOURCE_MAP_EXTENSION = '.srcmap.json'
FIND_BRANCHED_FROM = re.compile(r'Cr-Branched-From:.*master@\{#(\d+)\}')


def _add_components_from_dict(deps_dict, vars_dict, revisions_dict):
  """Add components from a dict representing a DEPS file."""
  if not deps_dict:
    # If the dictionary is None, bail out early.
    return

  for key, value in six.iteritems(deps_dict):
    url = rev = None
    if isinstance(value, str):
      url, _, rev = value.partition('@')
    elif isinstance(value, dict):
      if 'revision' in value:
        url = value['url']
        rev = value['revision']
      elif 'url' in value and value['url'] is not None:
        url, _, rev = value['url'].partition('@')

    if url and rev:
      url = url.format(**vars_dict)
      rev = rev.format(**vars_dict)
      revisions_dict[key] = {
          'name': _get_component_display_name(key),
          'rev': rev,
          'url': url
      }


def _clank_revision_file_to_revisions_dict(content):
  """Parse Clank revision file and return revisions dict."""
  component_revision_mappings = {}
  for line in content.splitlines():
    match = CLANK_REVISION_FILE_COMPONENT_REGEX.match(line)
    if not match:
      continue

    component = match.group(1)
    revision = match.group(2)
    component_revision_mappings[component] = revision

  if not component_revision_mappings:
    logs.log_error('Failed to get component revision mappings for clank.')
    return None

  chromium_revision = component_revision_mappings['chromium_revision']
  clank_revision = component_revision_mappings['clank_revision']

  # Initialize revisions dictionary with chromium repo.
  revisions_dict = get_component_revisions_dict(chromium_revision, None)
  if revisions_dict is None:
    logs.log_error(
        'Failed to get chromium component revisions.',
        chromium_revision=chromium_revision,
        clank_revision=clank_revision)
    return None

  # Add info on clank repo.
  revisions_dict['/src/clank'] = {
      'name': 'Clank',
      'url': CLANK_URL,
      'rev': clank_revision
  }
  return revisions_dict


def _get_component_display_name(name, default=None):
  """Display name for a component."""
  if default and name in ['', 'default', '/src']:
    return default.capitalize()

  names = name.split('/')
  name_index = -1
  if len(names) > 1 and names[-1] in COMPONENT_NAMES_BLACKLIST:
    # Skip the blacklisted names from right.
    name_index -= 1

  return names[name_index].capitalize()


def _get_display_revision(component_revision_dict):
  """Return display revision for a component revision dict."""
  if 'commit_pos' in component_revision_dict:
    return component_revision_dict['commit_pos']

  return component_revision_dict['rev'] or '<empty>'


def _get_link_text(start_component_revision_dict, end_component_revision_dict):
  """Return link text given a start and end revision. This is used in cases
  when revision url is not available."""
  start_revision = _get_display_revision(start_component_revision_dict)
  end_revision = _get_display_revision(end_component_revision_dict)

  if start_revision == end_revision:
    return str(start_revision)

  return '%s:%s' % (start_revision, end_revision)


def _get_link_url(start_component_revision_dict, end_component_revision_dict):
  """Return link text given a start and end revision. This is used in cases
  when revision url is not available."""
  url = start_component_revision_dict['url']
  if not url:
    return None

  vcs_viewer = source_mapper.get_vcs_viewer_for_url(url)
  if not vcs_viewer:
    # If we don't support the vcs yet, bail out.
    return None

  start_revision = _get_revision(start_component_revision_dict)
  end_revision = _get_revision(end_component_revision_dict)

  if start_revision == end_revision:
    return vcs_viewer.get_source_url_for_revision(start_revision)

  return vcs_viewer.get_source_url_for_revision_diff(start_revision,
                                                     end_revision)


def _get_revision(component_revision_dict):
  """Return revision for a component revision dict."""
  return component_revision_dict['rev']


def _get_url_content(url):
  """Read a potentially base64-encoded resource from the given URL."""
  if url.startswith(storage.GS_PREFIX):
    # Fetch a GCS path with authentication.
    url_data = storage.read_data(url)
    if url_data is None:
      return None

    url_content = url_data.decode('utf-8')
  else:
    # Fetch a regular url without authentication.
    url_content = utils.fetch_url(url)

    # Urls on googlesource.com return file data as base64 encoded to avoid
    # cross-site scripting attacks. If the requested url contains |format=text|,
    # then the output is base64 encoded. So, decode it first.
    if url_content and url.endswith('format=text'):
      url_content = base64.b64decode(url_content)

  return url_content


def _git_url_for_chromium_repository(repository):
  """Return git url for a chromium repository."""
  return '%s/%s.git' % (CHROMIUM_GIT_ROOT_URL, repository)


def _is_clank(url):
  """Return bool on whether this is a clank url or not."""
  # FIXME: Need a better way to check for this.
  return '/chrome-test-builds/android' in url


def _is_deps(url):
  """Return bool on whether this is a DEPS url or not."""
  return urllib.parse.urlparse(url).path.endswith('/DEPS')


def _src_map_to_revisions_dict(src_map, project_name):
  """Convert src map contents to revisions dict."""
  revisions_dict = {}

  for key in src_map:
    # Only add keys that have both url and rev attributes.
    if 'url' in src_map[key] and 'rev' in src_map[key]:
      revisions_dict[key] = {
          'name': _get_component_display_name(key, project_name),
          'rev': src_map[key]['rev'],
          'url': src_map[key]['url']
      }

  return revisions_dict


@memoize.wrap(memoize.FifoOnDisk(DISK_CACHE_SIZE))
@memoize.wrap(memoize.Memcache(60 * 60 * 24 * 30))  # 30 day TTL
def _git_commit_position_to_git_hash_for_chromium(revision, repository):
  """Return git hash for a git commit position using cr-rev.appspot.com."""
  request_variables = {
      'number': revision,
      'numbering_identifier': 'refs/heads/main',
      'numbering_type': 'COMMIT_POSITION',
      'project': 'chromium',
      'repo': repository,
      'fields': 'git_sha',
  }
  query_string = urllib.parse.urlencode(request_variables)
  query_url = '%s?%s' % (CRREV_NUMBERING_URL, query_string)
  url_content = _get_url_content(query_url)
  if url_content is None:
    logs.log_error('Failed to fetch git hash from url: ' + query_url)
    return None

  result_dict = _to_dict(url_content)
  if result_dict is None:
    logs.log_error('Failed to parse git hash from url: ' + query_url)
    return None

  return result_dict['git_sha']


def _to_dict(contents):
  """Parse |contents| as a dict, returning None on failure or if it's not a
  dict."""
  try:
    result = ast.literal_eval(contents)
    if isinstance(result, dict):
      return result

  except (ValueError, TypeError):
    pass

  return None


def deps_to_revisions_dict(content):
  """Parses DEPS content and returns a dictionary of revision variables."""
  local_context = {}
  global_context = {
      'Var': lambda x: local_context.get('vars', {}).get(x),
      'Str': str,
  }
  # pylint: disable=exec-used
  exec(content, global_context, local_context)

  revisions_dict = {}

  vars_dict = local_context.get('vars', {})
  deps_dict = local_context.get('deps')
  if not deps_dict:
    # |deps| variable is required. If it does not exist, we should raise an
    # exception.
    logs.log_error('Deps format has changed, code needs fixing.')
    return None
  _add_components_from_dict(deps_dict, vars_dict, revisions_dict)

  deps_os_dict = local_context.get('deps_os')
  if deps_os_dict:
    # |deps_os| variable is optional.
    for deps_os in list(deps_os_dict.values()):
      _add_components_from_dict(deps_os, vars_dict, revisions_dict)

  return revisions_dict


def get_components_list(component_revisions_dict, job_type):
  """Return a prioritized order of components based on job type."""
  components = sorted(component_revisions_dict.keys())

  if utils.is_chromium():
    # Components prioritization only applies to non-chromium projects.
    return components

  project_name = data_handler.get_project_name(job_type)
  if not project_name:
    # No project name found in job environment, return list as-is.
    return components

  main_repo = data_handler.get_main_repo(job_type)
  project_src = '/src/' + project_name
  for component in components.copy():
    if component_revisions_dict[component]['url'] == main_repo:
      # Matches recorded main repo.
      components.remove(component)
      components.insert(0, component)
      break

    if component == project_src:
      components.remove(component)
      components.insert(0, component)
      break

    if project_name.lower() in os.path.basename(component).lower():
      components.remove(component)
      components.insert(0, component)
      # Keep trying in case an exact match is found later.

  return components


def _get_revision_vars_url_format(job_type):
  """Return REVISION_VARS_URL from job environment if available. Otherwise,
  default to one set in project.yaml. For custom binary jobs, this is not
  applicable."""
  if job_type is None:
    # Force it to use env attribute in project.yaml.
    return local_config.ProjectConfig().get('env.REVISION_VARS_URL')

  custom_binary = data_handler.get_value_from_job_definition(
      job_type, 'CUSTOM_BINARY')
  if utils.string_is_true(custom_binary):
    return None

  rev_path = data_handler.get_value_from_job_definition_or_environment(
      job_type, 'REVISION_VARS_URL')
  rev_path = overrides.check_and_apply_overrides(
      rev_path, overrides.PLATFORM_ID_TO_REV_PATH_KEY)
  return rev_path


@memoize.wrap(memoize.FifoOnDisk(DISK_CACHE_SIZE))
@memoize.wrap(memoize.Memcache(60 * 60 * 24 * 30))  # 30 day TTL
def get_component_revisions_dict(revision, job_type):
  """Retrieve revision vars dict."""
  if revision == 0 or revision == '0' or revision is None:
    # Return empty dict for zero start revision.
    return {}

  revision_vars_url_format = _get_revision_vars_url_format(job_type)
  if not revision_vars_url_format:
    return None

  project_name = data_handler.get_project_name(job_type)
  revisions_dict = {}

  if utils.is_chromium():
    component = data_handler.get_component_name(job_type)
    repository = data_handler.get_repository_for_component(component)
    if repository and not _is_clank(revision_vars_url_format):
      revision_hash = _git_commit_position_to_git_hash_for_chromium(
          revision, repository)
      if revision_hash is None:
        return None

      # FIXME: While we check for this explicitly appended component in all
      # applicable cases that we know of within this codebase, if the dict
      # is shared with an external service (e.g. Predator) we may need to clean
      # this up beforehand.
      revisions_dict['/src'] = {
          'name': _get_component_display_name(component, project_name),
          'url': _git_url_for_chromium_repository(repository),
          'rev': revision_hash,
          'commit_pos': revision
      }

      # Use revision hash for info url later.
      revision = revision_hash

  revision_vars_url = revision_vars_url_format % revision
  url_content = _get_url_content(revision_vars_url)
  if not url_content:
    logs.log_error(
        'Failed to get component revisions from %s.' % revision_vars_url)
    return None

  # Parse as per DEPS format.
  if _is_deps(revision_vars_url):
    deps_revisions_dict = deps_to_revisions_dict(url_content)
    if not deps_revisions_dict:
      return None

    revisions_dict.update(deps_revisions_dict)
    return revisions_dict

  # Parse as per Clank DEPS format.
  if _is_clank(revision_vars_url):
    return _clank_revision_file_to_revisions_dict(url_content)

  # Default case: parse content as yaml.
  revisions_dict = _to_dict(url_content)
  if not revisions_dict:
    logs.log_error(
        'Failed to parse component revisions from %s.' % revision_vars_url)
    return None

  # Parse as per source map format.
  if revision_vars_url.endswith(SOURCE_MAP_EXTENSION):
    revisions_dict = _src_map_to_revisions_dict(revisions_dict, project_name)

  return revisions_dict


def get_component_list(revision, job_type):
  """Gets mapped revisions for a given revision."""
  return get_component_range_list(revision, revision, job_type)


def get_component_range_list(start_revision, end_revision, job_type):
  """Gets revision variable ranges for a changeset range."""
  start_component_revisions_dict = get_component_revisions_dict(
      start_revision, job_type)

  if start_revision == end_revision:
    end_component_revisions_dict = start_component_revisions_dict
  else:
    end_component_revisions_dict = get_component_revisions_dict(
        end_revision, job_type)

  if (start_component_revisions_dict is None or
      end_component_revisions_dict is None):
    return []

  component_revisions = []
  keys = get_components_list(end_component_revisions_dict, job_type)
  for key in keys:
    if not start_component_revisions_dict:
      # 0 start revision, can only show link text.
      end_component_display_revision = _get_display_revision(
          end_component_revisions_dict[key])
      component_name = end_component_revisions_dict[key]['name']
      component_revisions.append({
          'component': component_name,
          'link_text': '0:%s' % end_component_display_revision
      })
      continue

    if key not in start_component_revisions_dict:
      logs.log_warn('Key %s not found in start revision %s for job %s.' %
                    (key, start_revision, job_type))
      continue

    start_component_revision_dict = start_component_revisions_dict[key]
    end_component_revision_dict = end_component_revisions_dict[key]

    component_revisions.append({
        'component':
            start_component_revision_dict['name'],
        'link_text':
            _get_link_text(start_component_revision_dict,
                           end_component_revision_dict),
        'link_url':
            _get_link_url(start_component_revision_dict,
                          end_component_revision_dict)
    })

  return component_revisions


def get_start_and_end_revision(revision_range):
  """Return start and end revision for a regression range."""
  try:
    revision_range_list = revision_range.split(':')
    start_revision = int(revision_range_list[0])
    end_revision = int(revision_range_list[1])
  except:
    return [0, 0]

  return [start_revision, end_revision]


def format_revision_list(revisions, use_html=True):
  """Converts component revision list to html."""
  result = ''
  for revision in revisions:
    if revision['component']:
      result += '%s: ' % revision['component']

    if 'link_url' in revision and revision['link_url'] and use_html:
      result += '<a target="_blank" href="{link_url}">{link_text}</a>'.format(
          link_url=revision['link_url'], link_text=revision['link_text'])
    else:
      result += revision['link_text']

    if use_html:
      result += '<br />'
    else:
      result += '\n'

  return result


def convert_revision_to_integer(revision):
  """Returns an integer that represents the given revision."""
  # If revision is only decimal digits, like '249055', then do a simple
  # conversion.
  match = re.match(r'^\d+$', revision)
  if match:
    return int(revision)

  # If the revision has 4 parts separated by dots, like '34.0.1824.2', then do
  # the following conversion:
  # Pad the heads with up to 5 "0"s to allow them to be sorted properly, eg.:
  #   '34.0.1824.2'   -> 00034000000182400002
  #   '32.0.1700.107' -> 00032000000170000107
  # If neither of the two patterns matches, raise an error.
  match = re.match(r'^(\d{1,5})\.(\d{1,5})\.(\d{1,5})\.(\d{1,5})$', revision)
  if match:
    revision = '%s%s%s%s' % (match.group(1).zfill(5), match.group(2).zfill(5),
                             match.group(3).zfill(5), match.group(4).zfill(5))
    return int(revision)

  error = 'Unknown revision pattern: %s' % revision
  logs.log_error(error)
  raise ValueError(error)


def find_build_url(bucket_path, build_url_list, revision):
  """Returns the build url associated with a revision."""
  if not build_url_list:
    return None

  revision_pattern = revision_pattern_from_build_bucket_path(bucket_path)
  for build_url in build_url_list:
    match = re.match(revision_pattern, build_url)
    if not match:
      continue

    current_revision = convert_revision_to_integer(match.group(1))
    if current_revision == revision:
      return build_url

  return None


def find_min_revision_index(revisions_list, revision):
  """Find the min index for bisection. Find largest revision <= the given
  revision."""
  # bisect_left partitions |revisions_list| into 2 such that:
  #   all(val < revision for val in a[:index])
  #   all(val >= revision for val in a[index:])
  index = bisect.bisect_left(revisions_list, revision)
  if index < len(revisions_list) and revisions_list[index] == revision:
    return index

  if index > 0:
    return index - 1

  # No revisions <= given revision.
  return None


def find_max_revision_index(revisions_list, revision):
  """Find the max index for bisection. Find smallest revision >= the given
  revision."""
  index = bisect.bisect_left(revisions_list, revision)
  if index < len(revisions_list):
    return index

  # No revisions >= given revision.
  return None


def get_first_revision_in_list(revision_list):
  """Gets the first revision in list greater than or equal to MIN_REVISION."""
  first_revision = revision_list[0]

  min_revision = environment.get_value('MIN_REVISION')
  if not min_revision:
    return first_revision

  for revision in revision_list:
    if revision >= min_revision:
      return revision

  # No revision >= |MIN_REVISION| was found, store the error and just return
  # first revision.
  logs.log_error('Unable to find a revision >= MIN_REVISION.')
  return first_revision


def get_last_revision_in_list(revision_list):
  """Gets the last revision in list."""
  return revision_list[-1]


def get_real_revision(revision, job_type, display=False):
  """Convert the revision number into a real revision hash (e.g. git hash)."""
  if revision is None:
    # Bail early when caller passes revision from a non-existent attribute.
    return None

  component_revisions_dict = get_component_revisions_dict(revision, job_type)
  if not component_revisions_dict:
    return str(revision)

  keys = list(component_revisions_dict.keys())
  key = ('/src' if '/src' in keys else get_components_list(
      component_revisions_dict, job_type)[0])
  helper = _get_display_revision if display else _get_revision
  return helper(component_revisions_dict[key])


def needs_update(revision_file, revision):
  """Check a revision file against the provided revision
  to see if an update is required."""
  failure_wait_interval = environment.get_value('FAIL_WAIT')
  file_exists = False
  retry_limit = environment.get_value('FAIL_RETRIES')

  for _ in range(retry_limit):
    # NFS can sometimes return a wrong result on file existence, so redo
    # this check a couple of times to be sure.
    if not os.path.exists(revision_file):
      file_exists = False
      time.sleep(15)
      continue

    # Found the file, now try to read its contents.
    file_exists = True

    try:
      with open(revision_file, 'r') as file_handle:
        current_revision = file_handle.read()
    except:
      logs.log_error(
          'Error occurred while reading revision file %s.' % revision_file)
      time.sleep(utils.random_number(1, failure_wait_interval))
      continue

    if current_revision.isdigit():
      return int(revision) > int(current_revision)

    return str(revision) != str(current_revision)

  # If there is no revision file or if we have lost track of its revision,
  # then we do need to update the data bundle.
  if not file_exists:
    return True

  # An error has occurred and we have failed to read revision file
  # despite several retries. So, don't bother updating the data
  # bundle as it will probably fail as well.
  logs.log_error('Failed to read revision file, exiting.')
  return False


def write_revision_to_revision_file(revision_file, revision):
  """Writes a revision to the revision file."""
  try:
    with open(revision_file, 'wb') as file_handle:
      file_handle.write(str(revision).encode('utf-8'))
  except:
    logs.log_error(
        "Could not save revision to revision file '%s'" % revision_file)


def revision_pattern_from_build_bucket_path(bucket_path):
  """Get the revision pattern from a build bucket path."""
  return '.*?' + os.path.basename(bucket_path)


@memoize.wrap(memoize.FifoOnDisk(DISK_CACHE_SIZE))
@memoize.wrap(memoize.Memcache(60 * 60 * 24 * 30))  # 30 day TTL
def revision_to_branched_from(uri, revision):
  """Interrogates git code review server to find the branch-from
  revision of a component."""
  full_uri = "%s/+/%s?format=JSON" % (uri, revision)
  url_content = _get_url_content(full_uri)
  # gerrit intentionally returns nonsense in the first line.
  # See 'cross site script inclusion here:
  # https://gerrit-review.googlesource.com/Documentation/rest-api.html
  url_content = '\n'.join(url_content.splitlines()[1:])
  result = _to_dict(url_content)
  if not result:
    logs.log_error("Unable to retrieve and parse url: %s" % full_uri)
    return None
  msg = result.get('message', None)
  if not msg:
    logs.log_error("%s JSON had no 'message'" % full_uri)
    return None
  m = FIND_BRANCHED_FROM.search(msg)
  if not m:
    logs.log_error("%s JSON message lacked Cr-Branched-From" % full_uri)
    return None
  return m.group(1)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Handler for removing a testcase from a group."""

from flask import request

from clusterfuzz._internal.datastore import data_handler
from handlers import base_handler
from handlers.testcase_detail import show
from libs import handler
from libs import helpers


def remove_group(testcase_id):
  """Remove the testcase from a group."""
  testcase = helpers.get_testcase(testcase_id)
  group_id = testcase.group_id

  data_handler.remove_testcase_from_group(testcase)

  helpers.log(
      'Removed the testcase %s from the group %s' %
      (testcase.key.id(), group_id), helpers.MODIFY_OPERATION)

  return testcase


class Handler(base_handler.Handler):
  """Handler that removes a testcase from a group."""

  @handler.post(handler.JSON, handler.JSON)
  @handler.require_csrf_token
  @handler.check_admin_access
  def post(self):
    """Remove the issue from the testcase."""
    testcase_id = request.get('testcaseId')

    updated_testcase = remove_group(testcase_id)
    return self.render_json(show.get_testcase_detail(updated_testcase))



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests fuzzers.engine_common."""

import os
import unittest

import parameterized
from pyfakefs import fake_filesystem_unittest
import six

from clusterfuzz._internal.bot.fuzzers import engine_common
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.tests.test_libs import helpers as test_helpers
from clusterfuzz._internal.tests.test_libs import test_utils


class GetIssueOwnersTest(fake_filesystem_unittest.TestCase):
  """get_issue_owners tests."""

  def setUp(self):
    test_utils.set_up_pyfakefs(self)
    self.fs.create_dir('/test')

  def test_no_file(self):
    self.assertEqual([], engine_common.get_issue_owners('/test/does_not_exist'))

  def test_empty_file(self):
    self.fs.create_file('/test/fuzz_target.owners', contents='')
    self.assertEqual([], engine_common.get_issue_owners('/test/fuzz_target'))

  def test_all_allowed(self):
    self.fs.create_file('/test/fuzz_target.owners', contents='*')
    self.assertEqual([], engine_common.get_issue_owners('/test/fuzz_target'))

  def test_sections(self):
    """Test sections."""
    self.fs.create_file(
        '/test/fuzz_target.owners',
        contents=('file://ENG_REVIEW_OWNERS\n'
                  '  \n'
                  '#ignore me\n'
                  'dev1@example.com\n'
                  '\n'
                  'per-file PRESUBMIT*.py=dev2@chromium.org\n'
                  'dev3@example.com\n'
                  '  dev4@example.com    \n'))
    self.assertEqual(
        ['dev1@example.com', 'dev3@example.com', 'dev4@example.com'],
        engine_common.get_issue_owners('/test/fuzz_target'))
    self.assertEqual(
        ['dev1@example.com', 'dev3@example.com', 'dev4@example.com'],
        engine_common.get_issue_owners('/test/fuzz_target.exe'))


class GetIssueLabelsTest(fake_filesystem_unittest.TestCase):
  """get_issue_labels tests."""

  def setUp(self):
    test_utils.set_up_pyfakefs(self)
    self.fs.create_dir('/test')

  def test_no_file(self):
    """Test no labels file."""
    self.assertEqual([], engine_common.get_issue_labels('/test/does_not_exist'))

  def test_empty_file(self):
    """Test empty file."""
    self.fs.create_file('/test/fuzz_target.labels', contents='')
    self.assertEqual([], engine_common.get_issue_labels('/test/fuzz_target'))

  def test_well_formed(self):
    """Test well formed labels file."""
    self.fs.create_file(
        '/test/fuzz_target.labels', contents='label1\nlabel2\nlabel3\n')
    self.assertEqual(['label1', 'label2', 'label3'],
                     engine_common.get_issue_labels('/test/fuzz_target'))

  def test_empty_lines(self):
    """Test labels file with empty lines."""
    self.fs.create_file(
        '/test/fuzz_target.labels', contents='label1\n\t\nlabel2\n \nlabel3\n')
    self.assertEqual(['label1', 'label2', 'label3'],
                     engine_common.get_issue_labels('/test/fuzz_target'))


class GetTimeoutTestBase(unittest.TestCase):
  """Base class providing common functionality for TestCases that test the
  get_*_timeout() functions."""

  @staticmethod
  def function():
    raise NotImplementedError(
        'Children must override function() to use helper functions')

  def setUp(self):
    test_helpers.patch_environ(self)
    self.valid_hard_timeout = 500
    self.valid_merge_timeout = 15

  def _set_environment_values(self, environment_variable_values):
    """Set environment variable values based on the key, value pairs in
    |environment_variable_values|.
    """
    for key, value in six.iteritems(environment_variable_values):
      environment.set_value(key, value)

  def validation_helper(self, environment_variable_values):
    """Call self.function() and assert that it throws an AssertionError. Set
    environment variables based on |environment_variable_values| before calling
    self.function.
    """
    self._set_environment_values(environment_variable_values)
    with self.assertRaises(AssertionError):
      self.function()

  def call_helper(self, expected_value, environment_variable_values):
    """Test that self.function() == expected_value. Use
    |environment_variable_values| to mock environment.get_value before calling
    self.function."""
    self._set_environment_values(environment_variable_values)
    self.assertEqual(expected_value, self.function())


class GetHardTimeoutTest(GetTimeoutTestBase):
  """Get hard timeout tests."""
  # Use staticmethod to prevent python from passing self to
  # GetHardTimeoutTest.function().
  @staticmethod
  def function():
    return engine_common.get_hard_timeout()

  def test_hard_timeout_override_validation(self):
    """Test that get_hard_timeout rejects invalid values of
    HARD_TIMEOUT_OVERRIDE."""
    # Test that a negative HARD_TIMEOUT_OVERRIDE is rejected.
    self.validation_helper({
        'FUZZ_TEST_TIMEOUT': self.valid_hard_timeout,
        'HARD_TIMEOUT_OVERRIDE': -1
    })

  def test_hard_timeout_override_correctness(self):
    """Test that get_hard_timeout returns what we expect when we set
    HARD_TIMEOUT_OVERRIDE."""
    self.call_helper(self.valid_hard_timeout, {
        'HARD_TIMEOUT_OVERRIDE': self.valid_hard_timeout,
        'FUZZ_TEST_TIMEOUT': -1
    })

  def test_fuzz_test_timeout_correctness(self):
    """Test that get_hard_timeout returns what we expect when we set
    FUZZ_TEST_TIMEOUT."""
    self.call_helper(self.valid_hard_timeout,
                     {'FUZZ_TEST_TIMEOUT': self.valid_hard_timeout})


class GetMergeTimeoutTest(GetTimeoutTestBase):
  """Get merge timeout tests."""

  @staticmethod
  def function():
    return

  def test_validation(self):
    """Test that get_merge_timeout rejects invalid values of
    MERGE_TIMEOUT_OVERRIDE."""
    self._set_environment_values({
        'HARD_TIMEOUT_OVERRIDE': self.valid_hard_timeout,
        'MERGE_TIMEOUT_OVERRIDE': -1
    })

    with self.assertRaises(AssertionError):
      engine_common.get_merge_timeout(self.valid_merge_timeout)

  def test_correctness(self):
    """Test that get_merge_timeout returns what we expect."""
    # Basic sanity test.
    self.assertEqual(self.valid_merge_timeout,
                     engine_common.get_merge_timeout(self.valid_merge_timeout))

    # Test override works.
    override = 10
    self._set_environment_values({'MERGE_TIMEOUT_OVERRIDE': override})
    self.assertEqual(override,
                     engine_common.get_merge_timeout(self.valid_merge_timeout))


class FindFuzzerPathTest(fake_filesystem_unittest.TestCase):
  """find_fuzzer_path tests."""

  def setUp(self):
    test_utils.set_up_pyfakefs(self)
    test_helpers.patch_environ(self)
    self.build_dir = '/test'
    self.fuzzer_name = 'fuzz_target'
    self.fs.create_dir(self.build_dir)

  def _setup_fuzzer(self, fuzzer_name=None):
    """Create a file to represent the fuzzer."""
    if fuzzer_name is None:
      fuzzer_name = self.fuzzer_name
    fuzzer_path = os.path.join(self.build_dir, fuzzer_name)
    self.fs.create_file(fuzzer_path)
    return fuzzer_path

  def _find_fuzzer_path(self, fuzzer_name=None):
    """Return the result of a call to find_fuzzer_path with standard arguments.
    """
    if fuzzer_name is None:
      fuzzer_name = self.fuzzer_name
    return engine_common.find_fuzzer_path(self.build_dir, fuzzer_name)

  def test_finds_fuzzer(self):
    """Test common case of finding fuzzer in root of build directory."""
    self.assertEqual(self._setup_fuzzer(), self._find_fuzzer_path())

  def test_finds_fuzzer_with_legacy_prefix(self):
    """Test finding fuzzer, with legacy prefix."""
    environment.set_value('PROJECT_NAME', 'chromeos')
    self.assertEqual(self._setup_fuzzer(), self._find_fuzzer_path())

  def test_finds_fuzzer_with_legacy_prefix_in_name_and_env(self):
    """Test finding fuzzer, when legacy is set to the prefix in
    fuzzer_name."""
    environment.set_value('PROJECT_NAME', 'chromeos')
    fuzzer_name = 'chromeos_' + self.fuzzer_name
    self.assertEqual(
        self._setup_fuzzer(fuzzer_name), self._find_fuzzer_path(fuzzer_name))

  def test_only_finds_file(self):
    """Test that a directory is never returned."""
    self.fs.create_dir(os.path.join(self.build_dir, self.fuzzer_name))
    self.assertIsNone(self._find_fuzzer_path())

  def test_no_build_directory(self):
    """Test that no exception occurs when there is no build directory set."""
    self.assertIsNone(engine_common.find_fuzzer_path(None, self.fuzzer_name))


class GetStrategyProbabilityTest(unittest.TestCase):
  """Tests get_strategy_probability."""

  def setUp(self):
    test_helpers.patch_environ(self)

  def test_default_probability(self):
    """Test default probability is returned when |FUZZING_STRATEGIES| is not
    set."""
    self.assertEqual(0.33, engine_common.get_strategy_probability('foo', 0.33))

  def test_env_var_not_dict(self):
    """Test default probability is returned when |FUZZING_STRATEGIES| is not
     a dict."""
    environment.set_value('FUZZING_STRATEGIES', 'bad')
    self.assertEqual(0.33, engine_common.get_strategy_probability('foo', 0.33))

  def test_strategy_not_specified(self):
    """Test 0.0 weight is return when strategy is not defined in
    |FUZZING_STRATEGIES|."""
    environment.set_value('FUZZING_STRATEGIES',
                          '{"strategy_1": 0.5, "strategy_3": 0.3}')
    self.assertEqual(0.0,
                     engine_common.get_strategy_probability('strategy_2', 0.33))

  def test_strategy_specified(self):
    """Test weight is returned when strategy is defined in
    |FUZZING_STRATEGIES|."""
    environment.set_value('FUZZING_STRATEGIES',
                          '{"strategy_1": 0.5, "strategy_3": 1.0}')
    self.assertEqual(0.5,
                     engine_common.get_strategy_probability('strategy_1', 0.33))
    self.assertEqual(1.0,
                     engine_common.get_strategy_probability('strategy_3', 0.33))


class GetSeedCorpusPath(fake_filesystem_unittest.TestCase):
  """Tests get_seed_corpus_path."""
  FUZZ_TARGET_PATH = '/fuzz_target'
  FUZZ_TARGET_PATH_WITH_EXTENSION = FUZZ_TARGET_PATH + '.exe'

  def setUp(self):
    self.archive_path_without_extension = (
        self.FUZZ_TARGET_PATH + engine_common.SEED_CORPUS_ARCHIVE_SUFFIX)
    test_utils.set_up_pyfakefs(self)

  def _create_seed_corpus(self, extension):
    seed_corpus_archive_path = self.archive_path_without_extension + extension
    self.fs.create_file(seed_corpus_archive_path)
    return seed_corpus_archive_path

  def _get_seed_corpus_path(self, fuzz_target_path):
    return engine_common.get_seed_corpus_path(fuzz_target_path)

  @parameterized.parameterized.expand([
      ('.zip', FUZZ_TARGET_PATH),
      ('.tar.gz', FUZZ_TARGET_PATH),
      ('.zip', FUZZ_TARGET_PATH_WITH_EXTENSION),
      ('.tar.gz', FUZZ_TARGET_PATH_WITH_EXTENSION),
  ])
  def test_get_archive(self, extension, fuzz_target_path):
    """Tests that get_seed_corpus_path can find a seed corpus with
    |extension|."""
    self.assertEqual(
        self._create_seed_corpus(extension),
        self._get_seed_corpus_path(fuzz_target_path))

  @parameterized.parameterized.expand(
      [FUZZ_TARGET_PATH, FUZZ_TARGET_PATH_WITH_EXTENSION])
  def test_multiple_corpora(self, fuzz_target_path):
    """Tests that the function logs an error when target has multiple seed
    corpora."""
    test_helpers.patch(self, ['clusterfuzz._internal.metrics.logs.log_error'])
    self._create_seed_corpus('.tar.gz')
    self._create_seed_corpus('.zip')
    self.assertIsNotNone(self._get_seed_corpus_path(fuzz_target_path))
    self.assertEqual(self.mock.log_error.call_count, 1)

  @parameterized.parameterized.expand(
      [FUZZ_TARGET_PATH, FUZZ_TARGET_PATH_WITH_EXTENSION])
  def test_no_seed_corpus(self, fuzz_target_path):
    """Tests that the function returns None when target has no seed corpus."""
    self.assertIsNone(self._get_seed_corpus_path(fuzz_target_path))

  @parameterized.parameterized.expand(
      [FUZZ_TARGET_PATH, FUZZ_TARGET_PATH_WITH_EXTENSION])
  def test_invalid_extension(self, fuzz_target_path):
    """Tests that the function returns None when target has a seed corpus with
    an invalid file extension."""
    self._create_seed_corpus('.tar.invalid')
    self.assertIsNone(self._get_seed_corpus_path(fuzz_target_path))


class UnpackSeedCorpusIfNeededTest(fake_filesystem_unittest.TestCase):
  """Tests for unpack_seed_corpus_if_needed."""
  CORPUS_DIRECTORY = '/corpus'
  FUZZ_TARGET_PATH = '/fuzz_target'
  NUM_CORPUS_FILES = engine_common.MAX_FILES_FOR_UNPACK + 1

  def setUp(self):
    """Setup for unpack seed corpus if needed test."""
    self.data_path = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), 'data')

    zip_seed_corpus_path = os.path.join(self.data_path, 'seed_corpus.zip')
    with open(zip_seed_corpus_path, 'rb') as zip_seed_corpus_handle:
      self.zip_seed_corpus_contents = zip_seed_corpus_handle.read()

    targz_seed_corpus_path = os.path.join(self.data_path, 'seed_corpus.tar.gz')
    with open(targz_seed_corpus_path, 'rb') as targz_seed_corpus_handle:
      self.targz_seed_corpus_contents = targz_seed_corpus_handle.read()

    seed_corpus_with_subdirectories_path = os.path.join(
        self.data_path, 'seed_corpus_with_subdirectories.zip')
    with open(seed_corpus_with_subdirectories_path, 'rb') as seed_corpus_handle:
      self.seed_corpus_subdirs_contents = seed_corpus_handle.read()

    test_utils.set_up_pyfakefs(self)
    self.fs.create_dir(self.CORPUS_DIRECTORY)

  def _unpack_seed_corpus_if_needed(self, *args, **kwargs):
    return engine_common.unpack_seed_corpus_if_needed(
        self.FUZZ_TARGET_PATH, self.CORPUS_DIRECTORY, *args, **kwargs)

  def _list_corpus_dir(self):
    return os.listdir(self.CORPUS_DIRECTORY)

  def _assert_elements_equal(self, l1, l2):
    """Assert that the elements of |l1| and |l2| are equal. Modifies |l1| and
    |l2| by sorting them."""
    self.assertEqual(list(sorted(l1)), list(sorted(l2)))

  def _write_seed_corpus(self, data, extension):
    """Writes an archive seed corpus to the proper location."""
    seed_corpus_path = (
        self.FUZZ_TARGET_PATH + engine_common.SEED_CORPUS_ARCHIVE_SUFFIX +
        extension)

    with open(seed_corpus_path, 'wb+') as seed_corpus_handle:
      seed_corpus_handle.write(data)

  def _create_corpus_files(self, num_files):
    """Creates |num_files| files in the corpus directory."""
    filenames = []
    for file_num in range(num_files):
      file_num = str(file_num)
      self.fs.create_file(os.path.join(self.CORPUS_DIRECTORY, file_num))
      filenames.append(file_num)
    return filenames

  def test_no_seed_corpus(self):
    """Test that unpack_seed_corpus_if_needed does nothing when there is no seed
    corpus."""
    self._unpack_seed_corpus_if_needed()
    self._assert_elements_equal([], self._list_corpus_dir())

  def test_unpack_zip_seed_corpus_if_needed(self):
    """Tests unpack_seed_corpus_if_needed can unpack a seed corpus."""
    self._write_seed_corpus(self.zip_seed_corpus_contents, extension='.zip')
    expected_dir_contents = [
        '0000000000000000', '0000000000000001', '0000000000000002'
    ]
    self._unpack_seed_corpus_if_needed()
    self._assert_elements_equal(expected_dir_contents, self._list_corpus_dir())

  def test_unpack_targz_seed_corpus_if_needed(self):
    """Tests unpack_seed_corpus_if_needed can unpack a seed corpus."""
    self._write_seed_corpus(self.targz_seed_corpus_contents, '.tar.gz')
    expected_dir_contents = [
        '0000000000000000', '0000000000000001', '0000000000000002'
    ]
    self._unpack_seed_corpus_if_needed()
    self._assert_elements_equal(expected_dir_contents, self._list_corpus_dir())

  def test_max_files_for_unpack(self):
    """Tests unpack_seed_corpus_if_needed does not unpack a seed corpus if there
    are more than enough files in the corpus directory."""
    self._write_seed_corpus(self.zip_seed_corpus_contents, '.zip')
    corpus_files = self._create_corpus_files(self.NUM_CORPUS_FILES)
    self._unpack_seed_corpus_if_needed()
    self._assert_elements_equal(corpus_files, self._list_corpus_dir())

  def test_force_unpack(self):
    """Tests unpack_seed_corpus_if_needed unpacks a seed corpus even if there
    are more than files in the corpus directory, when the |force_unpack|
    argument is True."""
    self._write_seed_corpus(self.zip_seed_corpus_contents, '.zip')
    initial_corpus_files = self._create_corpus_files(self.NUM_CORPUS_FILES)
    seed_corpus_files = [
        '0000000000000000', '0000000000000001', '0000000000000002'
    ]
    expected_corpus_files = initial_corpus_files + seed_corpus_files
    self._unpack_seed_corpus_if_needed(force_unpack=True)
    self._assert_elements_equal(expected_corpus_files, self._list_corpus_dir())

  def test_unpack_seed_corpus_subdirs(self):
    """Test unpack_seed_corpus_if_needed method flattens a
    seed corpus with subdirectories."""
    expected_dir_contents = [
        '0000000000000000', '0000000000000001', '0000000000000002',
        '0000000000000003'
    ]
    self._write_seed_corpus(self.seed_corpus_subdirs_contents, '.zip')
    self._unpack_seed_corpus_if_needed()
    self._assert_elements_equal(expected_dir_contents, self._list_corpus_dir())


class GetRadamsaOutputFilenameTest(unittest.TestCase):
  """get_radamsa_output_filename tests."""

  def test_get_radamsa_output_filename(self):
    """Test get_radamsa_output_filename works as expected."""
    output_filename = engine_common.get_radamsa_output_filename('file', 0)
    self.assertEqual('radamsa-00001-file', output_filename)

  def test_no_double_prefix(self):
    """Test get_radamsa_output_filename strips an existing prefix before adding
    a new one."""
    output_filename = engine_common.get_radamsa_output_filename(
        'radamsa-00002-file', 0)
    self.assertEqual('radamsa-00001-file', output_filename)

  def test_filename_length_limit(self):
    """Test get_radamsa_output_filename does not return filenames that are too
    long."""
    filename_length_limit = 255
    output_filename = engine_common.get_radamsa_output_filename(
        filename_length_limit * 2 * 'a', 0)
    self.assertLessEqual(len(output_filename), filename_length_limit)


class ProcessSanitizerOptionsOverridesTest(fake_filesystem_unittest.TestCase):
  """process_sanitizer_options_overrides tests."""

  def setUp(self):
    test_utils.set_up_pyfakefs(self)
    self.fuzz_target = '/test'
    self.fuzz_target_options_file = self.fuzz_target + '.options'

  @parameterized.parameterized.expand([
      ('ASAN_OPTIONS', 'asan'),
      ('MSAN_OPTIONS', 'msan'),
      ('UBSAN_OPTIONS', 'ubsan'),
      ('HWASAN_OPTIONS', 'hwasan'),
  ])
  def test_sanitizer_options_changed(self, options_name, section_name):
    """Test that sanitizer options set in .options file are added to the
    environment variable."""
    environment.set_value(options_name, 'a=1:b=2:c=1')
    self.fs.create_file(
        self.fuzz_target_options_file,
        contents='[{section_name}]\nc=3:d=4'.format(section_name=section_name))
    engine_common.process_sanitizer_options_overrides(self.fuzz_target)
    self.assertEqual('a=1:b=2:c=3:d=4', environment.get_value(options_name))

  @parameterized.parameterized.expand([
      ('ASAN_OPTIONS', 'msan'),
      ('MSAN_OPTIONS', 'asan'),
      ('UBSAN_OPTIONS', 'msan'),
      ('HWASAN_OPTIONS', 'msan'),
  ])
  def test_sanitizer_options_not_changed_unrelated_section(
      self, options_name, section_name):
    """Test that sanitizer options are not changed when provided an unrelated
    sanitizer section name."""
    environment.set_value(options_name, 'a=1:b=2:c=1')
    self.fs.create_file(
        self.fuzz_target_options_file,
        contents='[{section_name}]\nc=3:d=4'.format(section_name=section_name))
    engine_common.process_sanitizer_options_overrides(self.fuzz_target)
    self.assertEqual('a=1:b=2:c=1', environment.get_value(options_name))

  @parameterized.parameterized.expand([
      ('ASAN_OPTIONS'),
      ('MSAN_OPTIONS'),
      ('UBSAN_OPTIONS'),
      ('HWASAN_OPTIONS'),
  ])
  def test_sanitizer_options_not_changed_no_options_file(self, options_name):
    """Test that sanitizer options are not changed and no exception occurs
    when .options file is not provided."""
    environment.set_value(options_name, 'a=1:b=2:c=1')
    engine_common.process_sanitizer_options_overrides(self.fuzz_target)
    self.assertEqual('a=1:b=2:c=1', environment.get_value(options_name))



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""ADB shell related functions."""

import collections
import os
import re
import signal
import socket
import subprocess
import tempfile
import threading
import time

from clusterfuzz._internal.base import persistent_cache
from clusterfuzz._internal.base import utils
from clusterfuzz._internal.metrics import logs
from clusterfuzz._internal.system import environment
from clusterfuzz._internal.system import shell

ADB_TIMEOUT = 1200  # Should be lower than |REBOOT_TIMEOUT|.
BAD_STATE_WAIT = 900
BOOT_WAIT_INTERVAL = 30
CUTTLEFISH_USER = 'vsoc-01'
CUTTLEFISH_CVD_PORT = 6520
DEFAULT_DEVICE_MEMORY_MB = 2048
DEVICE = collections.namedtuple('Device', ['serial', 'path'])
DEVICE_HANG_STRING = None
DEVICE_NOT_FOUND_STRING = 'error: device \'{serial}\' not found'
DEVICE_OFFLINE_STRING = 'error: device offline'
FACTORY_RESET_WAIT = 60
KERNEL_LOG_FILES = [
    '/proc/last_kmsg',
    '/sys/fs/pstore/console-ramoops',
]
MONKEY_PROCESS_NAME = 'monkey'
WAIT_FOR_DEVICE_TIMEOUT = 600
REBOOT_TIMEOUT = 3600
RECOVERY_CMD_TIMEOUT = 60
STOP_CVD_WAIT = 20

# Output patterns to parse "lsusb" output.
LSUSB_BUS_RE = re.compile(r'Bus\s+(\d+)\s+Device\s+(\d+):.*')
LSUSB_SERIAL_RE = re.compile(r'\s+iSerial\s+\d\s+(.*)')

# This is a constant value defined in usbdevice_fs.h in Linux system.
USBDEVFS_RESET = ord('U') << 8 | 20


def bad_state_reached():
  """Wait when device is in a bad state and exit."""
  persistent_cache.clear_values()
  logs.log_fatal_and_exit(
      'Device in bad state.', wait_before_exit=BAD_STATE_WAIT)


def connect_to_cuttlefish_device():
  """Connect to Cuttlefish cvd."""
  logs.log('Connect to cuttlefish device.')
  device_serial = environment.get_value('ANDROID_SERIAL')
  connect_cmd = f'{get_adb_path()} connect {device_serial}'
  return execute_command(connect_cmd, timeout=RECOVERY_CMD_TIMEOUT)


def copy_local_directory_to_remote(local_directory, remote_directory):
  """Copies local directory contents to a device directory."""
  create_directory_if_needed(remote_directory)
  if os.listdir(local_directory):
    run_command(['push', '%s/.' % local_directory, remote_directory])


def copy_local_file_to_remote(local_file_path, remote_file_path):
  """Copies local file to a device file."""
  create_directory_if_needed(os.path.dirname(remote_file_path))
  run_command(['push', local_file_path, remote_file_path])


def copy_remote_directory_to_local(remote_directory, local_directory):
  """Copies local directory contents to a device directory."""
  run_command(['pull', '%s/.' % remote_directory, local_directory])


def copy_remote_file_to_local(remote_file_path, local_file_path):
  """Copies device file to a local file."""
  shell.create_directory(
      os.path.dirname(local_file_path), create_intermediates=True)
  run_command(['pull', remote_file_path, local_file_path])


def create_directory_if_needed(device_directory):
  """Creates a directory on the device if it doesn't already exist."""
  run_shell_command(['mkdir', '-p', device_directory])


def directory_exists(directory_path):
  """Return whether a directory exists or not."""
  expected = '0'
  result = run_shell_command(
      '\'test -d "%s"; echo $?\'' % directory_path, log_error=False)
  return result == expected


def execute_command(cmd, timeout=None, log_error=True,
                    on_cuttlefish_host=False):
  """Spawns a subprocess to run the given shell command."""
  if on_cuttlefish_host and environment.is_android_cuttlefish():
    # Auto accept key fingerprint for ssh command.
    cmd = ('ssh -o StrictHostKeyChecking=no '
           f'{get_cuttlefish_ssh_target()} "{cmd}"')

  so = []
  # pylint: disable=consider-using-with
  output_dest = tempfile.TemporaryFile()
  # pylint: disable=subprocess-popen-preexec-fn,consider-using-with
  pipe = subprocess.Popen(
      cmd,
      executable='/bin/bash',
      stdout=output_dest,
      stderr=subprocess.STDOUT,
      shell=True,
      preexec_fn=lambda: signal.signal(signal.SIGPIPE, signal.SIG_DFL),
      bufsize=0)

  def run():
    """Thread target function that waits for subprocess to complete."""
    try:
      pipe.communicate()
      output_dest.seek(0)
      output = output_dest.read()
      output_dest.close()
      if output:
        so.append(output)
    except OSError as _:
      logs.log_warn('Failed to retrieve stdout from: %s' % cmd)
    if pipe.returncode:
      if log_error:
        logs.log_warn(
            '%s returned %d error code.' % (cmd, pipe.returncode),
            output=output)

  thread = threading.Thread(target=run)
  thread.start()
  thread.join(timeout)
  if thread.is_alive():
    logs.log_warn('Command %s timed out. Killing process.' % cmd)
    try:
      pipe.kill()
    except OSError:
      # Can't kill a dead process.
      pass

    return None

  bytes_output = b''.join(so)
  return bytes_output.strip().decode('utf-8', errors='ignore')


def copy_to_cuttlefish(src_path, dest_path, timeout=None):
  """Copy file to cuttlefish device."""
  cvd_address = get_cuttlefish_ssh_target()
  return execute_command(
      'scp -o StrictHostKeyChecking=no '
      f'-r {src_path} {cvd_address}:{dest_path}',
      timeout=timeout)


def factory_reset():
  """Reset device to factory state."""
  if environment.is_android_cuttlefish() or environment.is_android_emulator():
    # We cannot recover from this since there can be cases like userdata image
    # corruption in /data/data. Till the bug is fixed, we just need to wait
    # for reimage in next iteration.
    bad_state_reached()

  # A device can be stuck in a boot loop due to a bad clang library update.
  # Reverting that can bring a device back to good state.
  revert_asan_device_setup_if_needed()

  run_as_root()
  run_shell_command([
      'am', 'broadcast', '-a', 'android.intent.action.MASTER_CLEAR', '-n',
      'android/com.android.server.MasterClearReceiver'
  ])

  # Wait until the reset is complete.
  time.sleep(FACTORY_RESET_WAIT)


def file_exists(file_path):
  """Return whether a file exists or not."""
  expected = '0'
  result = run_shell_command(
      '\'test -f "%s"; echo $?\'' % file_path, log_error=False)
  return result == expected


def get_adb_command_line(adb_cmd):
  """Return adb command line for running an adb command."""
  device_serial = environment.get_value('ANDROID_SERIAL')
  adb_cmd_line = '%s -s %s %s' % (get_adb_path(), device_serial, adb_cmd)
  return adb_cmd_line


def get_adb_path():
  """Return path to ADB binary."""
  adb_path = environment.get_value('ADB')
  if adb_path:
    return adb_path

  return os.path.join(environment.get_platform_resources_directory(), 'adb')


def get_device_state():
  """Return the device status."""
  state_cmd = get_adb_command_line('get-state')
  return execute_command(state_cmd, timeout=RECOVERY_CMD_TIMEOUT)


def get_fastboot_command_line(fastboot_cmd):
  """Return fastboot command line for running a fastboot command."""
  fastboot_cmd_line = '%s %s' % (get_fastboot_path(), fastboot_cmd)
  return fastboot_cmd_line


def get_fastboot_path():
  """Return path to fastboot binary."""
  return os.path.join(environment.get_platform_resources_directory(),
                      'fastboot')


def get_file_checksum(file_path):
  """Return file's md5 checksum."""
  if not file_exists(file_path):
    return None

  return run_shell_command(['md5sum', '-b', file_path])


def get_file_size(file_path):
  """Return file's size."""
  if not file_exists(file_path):
    return None

  return int(run_shell_command(['stat', '-c%s', file_path]))


def get_kernel_log_content():
  """Return content of kernel logs."""
  kernel_log_content = ''
  for kernel_log_file in KERNEL_LOG_FILES:
    kernel_log_content += read_data_from_file(kernel_log_file) or ''

  return kernel_log_content


def get_ps_output():
  """Return ps output for all processes."""
  return run_shell_command(['ps', '-A'])


def get_process_and_child_pids(process_name):
  """Return process and child pids matching a process name."""
  pids = []
  ps_output_lines = get_ps_output().splitlines()

  while True:
    old_pids_length = len(pids)
    for line in ps_output_lines:
      data = line.split()

      # Make sure we have a valid pid and parent pid.
      try:
        # PID is in the second column.
        line_process_pid = int(data[1])
        # Parent PID is in the third column.
        line_parent_pid = int(data[2])
      except:
        continue

      # If we have already processed this pid, no more work to do.
      if line_process_pid in pids:
        continue

      # Process name is in the last column.
      # Monkey framework instances (if any) are children of our process launch,
      # so include these in pid list.
      line_process_name = data[-1]
      if (process_name in line_process_name or
          MONKEY_PROCESS_NAME in line_process_name):
        if process_name == line_process_name:
          pids.insert(0, line_process_pid)
        else:
          pids.append(line_process_pid)
        continue

      # Add child pids to end.
      if line_parent_pid in pids:
        pids.append(line_process_pid)

    new_pids_length = len(pids)
    if old_pids_length == new_pids_length:
      break

  return pids


def get_property(property_name):
  """Return property's value."""
  return run_shell_command(['getprop', property_name])


def hard_reset():
  """Perform a hard reset of the device."""
  if environment.is_android_cuttlefish() or environment.is_android_emulator():
    # There is no recovery step at this point for a cuttlefish bot, so just exit
    # and wait for reimage on next iteration.
    bad_state_reached()

  # For physical device.
  # Try hard-reset via sysrq-trigger (requires root).
  hard_reset_sysrq_cmd = get_adb_command_line(
      'shell echo b \\> /proc/sysrq-trigger')
  execute_command(hard_reset_sysrq_cmd, timeout=RECOVERY_CMD_TIMEOUT)

  # Try soft-reset now (does not require root).
  soft_reset_cmd = get_adb_command_line('reboot')
  execute_command(soft_reset_cmd, timeout=RECOVERY_CMD_TIMEOUT)


def kill_processes_and_children_matching_name(process_name):
  """Kills process along with children matching names."""
  process_and_child_pids = get_process_and_child_pids(process_name)
  if not process_and_child_pids:
    return

  kill_command = ['kill', '-9'] + process_and_child_pids
  run_shell_command(kill_command)


def read_data_from_file(file_path):
  """Return device's file content."""
  if not file_exists(file_path):
    return None

  return run_shell_command(['cat', '"%s"' % file_path])


def reboot():
  """Reboots device."""
  run_command('reboot')


def start_cuttlefish_device(use_kernel=False):
  """Start the cuttlefish device."""
  cvd_dir = environment.get_value('CVD_DIR')
  cvd_bin_dir = os.path.join(cvd_dir, 'bin')
  launch_cvd_path = os.path.join(cvd_bin_dir, 'launch_cvd')

  device_memory_mb = environment.get_value('DEVICE_MEMORY_MB',
                                           DEFAULT_DEVICE_MEMORY_MB)
  launch_cvd_command_line = (
      f'{launch_cvd_path} -daemon -memory_mb {device_memory_mb} '
      '-report_anonymous_usage_stats Y')
  if use_kernel:
    kernel_path = os.path.join(cvd_dir, 'bzImage')
    initramfs_path = os.path.join(cvd_dir, 'initramfs.img')
    launch_cvd_command_line += (
        f' -kernel_path={kernel_path} -initramfs_path={initramfs_path}')

  execute_command(launch_cvd_command_line, on_cuttlefish_host=True)


def stop_cuttlefish_device():
  """Stops the cuttlefish device."""
  cvd_dir = environment.get_value('CVD_DIR')
  cvd_bin_dir = os.path.join(cvd_dir, 'bin')
  stop_cvd_cmd = os.path.join(cvd_bin_dir, 'stop_cvd')

  execute_command(
      stop_cvd_cmd, timeout=RECOVERY_CMD_TIMEOUT, on_cuttlefish_host=True)
  time.sleep(STOP_CVD_WAIT)


def restart_cuttlefish_device():
  """Restarts the cuttlefish device."""
  cvd_dir = environment.get_value('CVD_DIR')
  cvd_bin_dir = os.path.join(cvd_dir, 'bin')
  restart_cvd_cmd = os.path.join(cvd_bin_dir, 'restart_cvd')

  execute_command(restart_cvd_cmd, on_cuttlefish_host=True)


def recreate_cuttlefish_device():
  """Recreate cuttlefish device, restoring from backup images."""
  logs.log('Reimaging cuttlefish device.')
  cvd_dir = environment.get_value('CVD_DIR')

  stop_cuttlefish_device()

  # Delete all existing images.
  rm_cmd = f'rm -rf {cvd_dir}/*'
  execute_command(rm_cmd, timeout=RECOVERY_CMD_TIMEOUT, on_cuttlefish_host=True)

  # Copy and Combine cvd host package and OTA images.
  image_directory = environment.get_value('IMAGES_DIR')
  for image_filename in os.listdir(image_directory):
    if image_filename.endswith('.zip') or image_filename.endswith('.tar.gz'):
      continue
    image_src = os.path.join(image_directory, image_filename)
    image_dest = os.path.join(cvd_dir, image_filename)
    copy_to_cuttlefish(image_src, image_dest)

  start_cuttlefish_device()


def remount():
  """Remount /system as read/write."""
  run_as_root()
  run_command('remount')
  wait_for_device()
  run_as_root()


def remove_directory(device_directory, recreate=False):
  """Delete everything inside of a device directory and recreate if needed."""
  run_shell_command('rm -rf %s' % device_directory, root=True)
  if recreate:
    create_directory_if_needed(device_directory)


def remove_file(file_path):
  """Remove file."""
  run_shell_command('rm -f %s' % file_path, root=True)


def reset_device_connection():
  """Reset the connection to the physical device through USB. Returns whether
  or not the reset succeeded."""
  if environment.is_android_cuttlefish():
    restart_cuttlefish_device()
  else:
    # Physical device. Try restarting usb.
    reset_usb()

  # Check device status.
  state = get_device_state()
  if state != 'device':
    logs.log_warn('Device state is %s, unable to recover using usb reset/'
                  'cuttlefish reconnect.' % str(state))
    return False

  return True


def get_cuttlefish_device_ip():
  """Return the ip address of cuttlefish device."""
  try:
    return socket.gethostbyname('cuttlefish')
  except socket.gaierror:
    logs.log_fatal_and_exit('Unable to get cvd ip address on cuttlefish host.')
  return None


def set_cuttlefish_device_serial():
  """Set the ANDROID_SERIAL to cuttlefish ip and port."""
  device_serial = '%s:%d' % (get_cuttlefish_device_ip(), CUTTLEFISH_CVD_PORT)
  environment.set_value('ANDROID_SERIAL', device_serial)
  logs.log('Set cuttlefish device serial: %s' % device_serial)


def get_cuttlefish_ssh_target():
  """Return the target for cuttlefish ssh connection."""
  return f'{CUTTLEFISH_USER}@{get_cuttlefish_device_ip()}'


def get_device_path():
  """Gets a device path to be cached and used by reset_usb."""

  def _get_usb_devices():
    """Returns a list of device objects containing a serial and USB path."""
    usb_list_cmd = 'lsusb -v'
    output = execute_command(usb_list_cmd, timeout=RECOVERY_CMD_TIMEOUT)
    if output is None:
      logs.log_error('Failed to populate usb devices using lsusb, '
                     'host restart might be needed.')
      bad_state_reached()

    devices = []
    path = None
    for line in output.splitlines():
      match = LSUSB_BUS_RE.match(line)
      if match:
        path = '/dev/bus/usb/%s/%s' % (match.group(1), match.group(2))
        continue

      match = LSUSB_SERIAL_RE.match(line)
      if path and match and match.group(1):
        serial = match.group(1)
        devices.append(DEVICE(serial, path))

    return devices

  def _get_device_path_for_serial():
    """Return device path. Assumes a simple ANDROID_SERIAL."""
    devices = _get_usb_devices()
    for device in devices:
      if device_serial == device.serial:
        return device.path

    return None

  def _get_device_path_for_usb():
    """Returns a device path.

    Assumes ANDROID_SERIAL in the form "usb:<identifier>"."""
    # Android serial may reference a usb device rather than a serial number.
    device_id = device_serial[len('usb:'):]
    bus_number = int(
        open('/sys/bus/usb/devices/%s/busnum' % device_id).read().strip())
    device_number = int(
        open('/sys/bus/usb/devices/%s/devnum' % device_id).read().strip())
    return '/dev/bus/usb/%03d/%03d' % (bus_number, device_number)

  if environment.is_android_cuttlefish():
    return None

  device_serial = environment.get_value('ANDROID_SERIAL')
  if device_serial.startswith('usb:'):
    return _get_device_path_for_usb()

  return _get_device_path_for_serial()


def reset_usb():
  """Reset USB bus for a device serial."""
  if environment.is_android_cuttlefish() or environment.is_android_emulator():
    # Nothing to do here.
    return True

  # App Engine does not let us import this.
  import fcntl

  # We need to get latest device path since it could be changed in reboots or
  # adb root restarts.
  try:
    device_path = get_device_path()
  except IOError:
    # We may reach this state if the device is no longer available.
    device_path = None

  if not device_path:
    # Try pulling from cache (if available).
    device_path = environment.get_value('DEVICE_PATH')
  if not device_path:
    logs.log_warn('No device path found, unable to reset usb.')
    return False

  try:
    with open(device_path, 'w') as f:
      fcntl.ioctl(f, USBDEVFS_RESET)
  except:
    logs.log_warn('Failed to reset usb.')
    return False

  # Wait for usb to recover.
  wait_for_device(recover=False)
  return True


def revert_asan_device_setup_if_needed():
  """Reverts ASan device setup if installed."""
  if not environment.get_value('ASAN_DEVICE_SETUP'):
    return

  device_id = environment.get_value('ANDROID_SERIAL')
  device_argument = '--device %s' % device_id
  revert_argument = '--revert'
  asan_device_setup_script_path = os.path.join(
      environment.get_platform_resources_directory(), 'third_party',
      'asan_device_setup.sh')

  command = '%s %s %s' % (asan_device_setup_script_path, device_argument,
                          revert_argument)
  execute_command(command, timeout=RECOVERY_CMD_TIMEOUT)


def run_as_root():
  """Restart adbd and runs as root."""
  # Check if we are already running as root. If yes bail out.
  if get_property('service.adb.root') == '1':
    return

  wait_for_device()
  run_command('root')
  wait_for_device()


def run_command(cmd,
                log_output=False,
                log_error=True,
                timeout=None,
                recover=True):
  """Run a command in adb shell."""
  if isinstance(cmd, list):
    cmd = ' '.join([str(i) for i in cmd])
  if log_output:
    logs.log('Running: adb %s' % cmd)
  if not timeout:
    timeout = ADB_TIMEOUT

  output = execute_command(get_adb_command_line(cmd), timeout, log_error)
  if not recover or environment.is_android_emulator():
    if log_output:
      logs.log('Output: (%s)' % output)
    return output

  device_not_found_string_with_serial = DEVICE_NOT_FOUND_STRING.format(
      serial=environment.get_value('ANDROID_SERIAL'))
  if (output in [
      DEVICE_HANG_STRING, DEVICE_OFFLINE_STRING,
      device_not_found_string_with_serial
  ]):
    logs.log_warn('Unable to query device, resetting device connection.')
    if reset_device_connection():
      # Device has successfully recovered, re-run command to get output.
      # Continue execution and validate output next for |None| condition.
      output = execute_command(get_adb_command_line(cmd), timeout, log_error)
    else:
      output = DEVICE_HANG_STRING

  if output is DEVICE_HANG_STRING:
    # Handle the case where our command execution hung. This is usually when
    # device goes into a bad state and only way to recover is to restart it.
    logs.log_warn('Unable to query device, restarting device to recover.')
    hard_reset()

    # Wait until we've booted and try the command again.
    wait_until_fully_booted()
    output = execute_command(get_adb_command_line(cmd), timeout, log_error)

  if log_output:
    logs.log('Output: (%s)' % output)
  return output


def run_shell_command(cmd,
                      log_output=False,
                      log_error=True,
                      root=False,
                      timeout=None,
                      recover=True):
  """Run adb shell command (with root if needed)."""

  def _escape_specials(command):
    return command.replace('\\', '\\\\').replace('"', '\\"')

  if isinstance(cmd, list):
    cmd = ' '.join([str(i) for i in cmd])

  if cmd[0] not in ['"', "'"]:
    cmd = '"{}"'.format(_escape_specials(cmd))

  if root:
    root_cmd_prefix = 'su root sh -c '

    # The arguments to adb shell need to be quoted, so if we're using
    # su root sh -c, quote the combined command
    full_cmd = 'shell \'{}{}\''.format(root_cmd_prefix, cmd)
  else:
    full_cmd = 'shell {}'.format(cmd)

  return run_command(
      full_cmd,
      log_output=log_output,
      log_error=log_error,
      timeout=timeout,
      recover=recover)


def run_fastboot_command(cmd, log_output=True, log_error=True, timeout=None):
  """Run a command in fastboot shell."""
  if environment.is_android_cuttlefish():
    # We can't run fastboot commands on Android cuttlefish instances.
    return None

  if isinstance(cmd, list):
    cmd = ' '.join([str(i) for i in cmd])
  if log_output:
    logs.log('Running: fastboot %s' % cmd)
  if not timeout:
    timeout = ADB_TIMEOUT

  output = execute_command(get_fastboot_command_line(cmd), timeout, log_error)
  return output


def setup_adb():
  """Sets up ADB binary for use."""
  adb_binary_path = get_adb_path()

  # Make sure that ADB env var is set.
  if not environment.get_value('ADB'):
    environment.set_value('ADB', adb_binary_path)


def start_shell():
  """Stops shell."""
  # Make sure we are running as root.
  run_as_root()

  run_shell_command('start')
  wait_until_fully_booted()


def stop_shell():
  """Stops shell."""
  # Make sure we are running as root.
  run_as_root()
  run_shell_command('stop')


def time_since_last_reboot():
  """Return time in seconds since last reboot."""
  uptime_string = run_shell_command(['cat', '/proc/uptime']).split(' ')[0]
  try:
    return float(uptime_string)
  except ValueError:
    # Sometimes, adb can just hang or return null output. In these cases, just
    # return infinity uptime value.
    return float('inf')


def wait_for_device(recover=True):
  """Waits indefinitely for the device to come online."""
  run_command(
      'wait-for-device', timeout=WAIT_FOR_DEVICE_TIMEOUT, recover=recover)


def wait_until_fully_booted():
  """Wait until device is fully booted or timeout expires."""

  def boot_completed():
    expected = '1'
    result = run_shell_command('getprop sys.boot_completed', log_error=False)
    return result == expected

  def drive_ready():
    expected = '0'
    result = run_shell_command('\'test -d "/"; echo $?\'', log_error=False)
    return result == expected

  def package_manager_ready():
    expected = 'package:/system/framework/framework-res.apk'
    result = run_shell_command('pm path android', log_error=False)
    if not result:
      return False

    # Ignore any extra messages before or after the result we want.
    return expected in result.splitlines()

  # Make sure we are not already recursing inside this function.
  if utils.is_recursive_call():
    return False

  # Wait until device is online.
  wait_for_device()

  start_time = time.time()

  is_boot_completed = False
  is_drive_ready = False
  is_package_manager_ready = False
  while time.time() - start_time < REBOOT_TIMEOUT:
    # TODO(mbarbella): Investigate potential optimizations.
    # The package manager check should also work for shell restarts.
    if not is_drive_ready:
      is_drive_ready = drive_ready()
    if not is_package_manager_ready:
      is_package_manager_ready = package_manager_ready()
    if not is_boot_completed:
      is_boot_completed = boot_completed()

    if is_drive_ready and is_package_manager_ready and is_boot_completed:
      return True

    time.sleep(BOOT_WAIT_INTERVAL)

  factory_reset()
  logs.log_fatal_and_exit(
      'Device failed to finish boot. Reset to factory settings and exited.')

  # Not reached.
  return False


def write_command_line_file(command_line, app_path):
  """Write command line file with command line argument for the application."""
  command_line_path = environment.get_value('COMMAND_LINE_PATH')
  if not command_line_path:
    return

  # Algorithm for filtering current command line.
  # 1. Remove |APP_PATH| from front.
  # 2. Add 'chrome ' to start.
  # 3. Strip for whitespaces at start and end.
  command_line_without_app_path = command_line.replace('%s ' % app_path, '')
  command_line_file_contents = 'chrome %s' % (
      command_line_without_app_path.strip())

  write_data_to_file(command_line_file_contents, command_line_path)


def write_data_to_file(contents, file_path):
  """Writes content to file."""
  # If this is a file in /system, we need to remount /system as read-write and
  # after file is written, revert it back to read-only.
  is_system_file = file_path.startswith('/system')
  if is_system_file:
    remount()

  # Write file with desired contents.
  run_shell_command("\"echo -n '%s' | su root dd of=%s\"" % (contents.replace(
      '"', '\\"'), file_path))

  # Make command line file is readable.
  run_shell_command('chmod 0644 %s' % file_path, root=True)

  if is_system_file:
    reboot()
    wait_until_fully_booted()



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Gestures related functions."""

import os
import random

from . import adb

# Fixed delay in milliseconds between consecutive monkey events.
MONKEY_THROTTLE_DELAY = 100

# Maximum number of monkey events per testcase.
NUM_MONKEY_EVENTS = 25


def get_random_gestures(_):
  """Return a random gesture seed from monkey framework natively supported by
  Android OS."""
  random_seed = random.getrandbits(32)
  gesture = 'monkey,%s' % str(random_seed)
  return [gesture]


def run_gestures(gestures, *_):
  """Run the provided interaction gestures."""
  package_name = os.getenv('PKG_NAME')
  if not package_name:
    # No package to send gestures to, bail out.
    return

  if len(gestures) != 1 or not gestures[0].startswith('monkey'):
    # Bad gesture string, bail out.
    return

  monkey_seed = gestures[0].split(',')[-1]
  adb.run_shell_command([
      'monkey', '-p', package_name, '-s', monkey_seed, '--throttle',
      str(MONKEY_THROTTLE_DELAY), '--ignore-security-exceptions',
      str(NUM_MONKEY_EVENTS)
  ])



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for monorail issue management."""

import datetime
import unittest

import mock
import six

from clusterfuzz._internal.tests.test_libs import helpers
from libs.issue_management import monorail
from libs.issue_management.monorail import issue_tracker_manager
from libs.issue_management.monorail.comment import Comment as MonorailComment
from libs.issue_management.monorail.issue import Issue as MonorailIssue


class IssueTrackerManager(issue_tracker_manager.IssueTrackerManager):
  """Mock issue tracker manager."""

  def __init__(self, project_name, mock_issues):
    # pylint: disable=super-init-not-called
    self.project_name = project_name
    self.last_issue = None
    self.mock_issues = mock_issues

  def get_issue(self, issue_id):
    issue = self.mock_issues.get(issue_id)
    if not issue:
      return None

    issue.itm = self
    return issue

  def save(self, issue, *args, **kwargs):  # pylint: disable=unused-argument,arguments-differ
    self.last_issue = issue


class MonorailTests(unittest.TestCase):
  """Tests for the monorail issue tracker."""

  def setUp(self):
    helpers.patch(self, [
        'libs.issue_management.monorail.issue_tracker_manager.'
        'IssueTrackerManager.get_issues',
    ])

    mock_issue = MonorailIssue()
    mock_issue.id = 1337
    mock_issue.summary = 'summary'
    mock_issue.body = 'body'
    mock_issue.owner = 'owner'
    mock_issue.reporter = 'reporter'
    mock_issue.status = 'New'
    mock_issue.add_label('label1')
    mock_issue.add_label('label2')
    mock_issue.add_component('A>B')
    mock_issue.add_component('C>D')
    mock_issue.add_cc('cc@cc.com')

    mock_comment0 = MonorailComment()
    mock_comment0.author = 'author'
    mock_comment0.cc = ['-removed@cc.com', 'cc@cc.com']
    mock_comment0.labels = ['-label0', 'label1']
    mock_comment0.components = ['-E>F', 'A>B']
    mock_comment0.comment = 'comment'
    mock_comment0.summary = 'summary'
    mock_comment0.status = 'status'
    mock_comment0.owner = 'owner'

    mock_comment1 = MonorailComment()
    mock_comment1.author = 'author'
    mock_comment1.comment = 'comment'

    mock_issue.comments = [
        mock_comment0,
        mock_comment1,
    ]

    mock_issue_merged = MonorailIssue()
    mock_issue_merged.id = 1338
    mock_issue_merged.merged_into = 1337
    mock_issue_merged.merged_into_project = 'project'
    mock_issue_merged.closed = datetime.datetime(2019, 1, 1)

    mock_issue_merged_another_project = MonorailIssue()
    mock_issue_merged_another_project.id = 1339
    mock_issue_merged_another_project.merged_into = 1
    mock_issue_merged_another_project.merged_into_project = 'different-project'

    mock_issues = {
        1337: mock_issue,
        1338: mock_issue_merged,
        1339: mock_issue_merged_another_project,
    }

    self.itm = IssueTrackerManager('project', mock_issues)
    self.issue_tracker = monorail.IssueTracker(self.itm)

  def test_get_issue(self):
    """Test get_issue."""
    self.assertIsNone(self.issue_tracker.get_issue(1))

    issue = self.issue_tracker.get_issue(1337)
    self.assertEqual(1337, issue.id)
    self.assertEqual('summary', issue.title)
    self.assertEqual('body', issue.body)
    self.assertEqual('owner', issue.assignee)
    self.assertEqual('reporter', issue.reporter)
    self.assertEqual('New', issue.status)
    self.assertIsNone(issue.merged_into)

    six.assertCountEqual(self, [
        'label1',
        'label2',
    ], issue.labels)
    six.assertCountEqual(self, [
        'A>B',
        'C>D',
    ], issue.components)
    six.assertCountEqual(self, [
        'cc@cc.com',
    ], issue.ccs)

    issue = self.issue_tracker.get_issue(1338)
    self.assertEqual(1338, issue.id)
    self.assertEqual(datetime.datetime(2019, 1, 1), issue.closed_time)

  def test_new_issue(self):
    """Test new_issue."""
    issue = self.issue_tracker.new_issue()
    issue.assignee = 'owner'
    issue.title = 'summary'
    issue.body = 'body'
    issue.assignee = 'owner'
    issue.reporter = 'reporter'
    issue.status = 'New'
    issue.labels.add('label1')
    issue.labels.add('label2')
    issue.components.add('A>B')
    issue.components.add('C>D')
    issue.ccs.add('cc@cc.com')
    issue.save(new_comment='comment')

    monorail_issue = self.itm.last_issue

    self.assertEqual('summary', monorail_issue.summary)
    self.assertEqual('body', monorail_issue.body)
    self.assertEqual('owner', monorail_issue.owner)
    self.assertEqual('reporter', monorail_issue.reporter)
    self.assertEqual('New', monorail_issue.status)
    self.assertEqual('comment', monorail_issue.comment)

    six.assertCountEqual(self, [
        'label1',
        'label2',
    ], monorail_issue.labels)
    six.assertCountEqual(self, [
        'A>B',
        'C>D',
    ], monorail_issue.components)
    six.assertCountEqual(self, [
        'cc@cc.com',
    ], monorail_issue.cc)

  def test_actions(self):
    """Test actions."""
    issue = self.issue_tracker.get_issue(1337)
    actions = list(issue.actions)
    self.assertEqual(2, len(actions))

    self.assertEqual('summary', actions[0].title)
    self.assertEqual('comment', actions[0].comment)
    self.assertEqual('owner', actions[0].assignee)
    self.assertEqual('status', actions[0].status)
    six.assertCountEqual(self, ['cc@cc.com'], actions[0].ccs.added)
    six.assertCountEqual(self, ['removed@cc.com'], actions[0].ccs.removed)
    six.assertCountEqual(self, ['label1'], actions[0].labels.added)
    six.assertCountEqual(self, ['label0'], actions[0].labels.removed)
    six.assertCountEqual(self, ['A>B'], actions[0].components.added)
    six.assertCountEqual(self, ['E>F'], actions[0].components.removed)

    self.assertIsNone(actions[1].title)
    self.assertEqual('comment', actions[1].comment)
    self.assertIsNone(actions[1].assignee)
    self.assertIsNone(actions[1].status)
    six.assertCountEqual(self, [], actions[1].ccs.added)
    six.assertCountEqual(self, [], actions[1].ccs.removed)
    six.assertCountEqual(self, [], actions[1].labels.added)
    six.assertCountEqual(self, [], actions[1].labels.removed)
    six.assertCountEqual(self, [], actions[1].components.added)
    six.assertCountEqual(self, [], actions[1].components.removed)

  def test_modify_labels(self):
    """Test modifying labels."""
    issue = self.issue_tracker.get_issue(1337)
    issue.labels.add('Label3')
    issue.labels.remove('laBel1')
    six.assertCountEqual(self, ['label2', 'Label3'], issue.labels)
    issue.save()

    six.assertCountEqual(self, ['label2', 'Label3', '-laBel1'],
                         self.itm.last_issue.labels)

  def test_modify_components(self):
    """Test modifying labels."""
    issue = self.issue_tracker.get_issue(1337)
    issue.components.add('Y>Z')
    issue.components.remove('a>B')
    six.assertCountEqual(self, ['C>D', 'Y>Z'], issue.components)
    issue.save()

    six.assertCountEqual(self, ['-a>B', 'C>D', 'Y>Z'],
                         self.itm.last_issue.components)

  def test_get_original_issue(self):
    """Test get_original_issue."""
    issue = self.issue_tracker.get_original_issue(1338)
    self.assertEqual(1337, issue.id)

  def test_find_issues(self):
    """Test find_issues."""
    issue0 = MonorailIssue()
    issue0.id = 1

    issue1 = MonorailIssue()
    issue1.id = 2
    self.mock.get_issues.return_value = [
        issue0,
        issue1,
    ]
    issues = self.issue_tracker.find_issues(
        keywords=['one', 'two'], only_open=True)
    six.assertCountEqual(self, [1, 2], [issue.id for issue in issues])

    self.mock.get_issues.assert_has_calls([
        mock.call(mock.ANY, '"one" "two"', can='open'),
    ])

    issues = self.issue_tracker.find_issues(
        keywords=['one', 'two'], only_open=False)
    six.assertCountEqual(self, [1, 2], [issue.id for issue in issues])

    self.mock.get_issues.assert_has_calls([
        mock.call(mock.ANY, '"one" "two"', can='all'),
    ])

  def test_find_issues_url(self):
    """Test find_issues_url."""
    url = self.issue_tracker.find_issues_url(
        keywords=['one', 'two'], only_open=False)
    self.assertEqual(
        'https://bugs.chromium.org/p/project/issues/list'
        '?can_id=1&q=%22one%22+%22two%22', url)

    url = self.issue_tracker.find_issues_url(
        keywords=['one', 'two'], only_open=True)
    self.assertEqual(
        'https://bugs.chromium.org/p/project/issues/list'
        '?can_id=2&q=%22one%22+%22two%22', url)

  def test_issue_url(self):
    """Test issue_url."""
    issue_url = self.issue_tracker.issue_url(1337)
    self.assertEqual(
        'https://bugs.chromium.org/p/project/issues/detail?id=1337', issue_url)

  def test_merged_into_different_project(self):
    """Test merged_into for a different issue tracker project."""
    issue = self.issue_tracker.get_issue(1339)
    self.assertIsNone(issue.merged_into)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Content viewer."""

from flask import request

from clusterfuzz._internal.base import utils
from clusterfuzz._internal.google_cloud_utils import blobs
from handlers import base_handler
from libs import access
from libs import handler
from libs import helpers

MAX_ALLOWED_CONTENT_SIZE = 10 * 1024 * 1024


class Handler(base_handler.Handler):
  """Content Viewer."""

  @handler.get(handler.HTML)
  def get(self):
    """Get the HTML page."""
    key = request.get('key')
    if not key:
      raise helpers.EarlyExitException('No key provided.', 400)

    testcase_id = request.get('testcase_id')
    if testcase_id:
      testcase = helpers.get_testcase(testcase_id)
      if not access.can_user_access_testcase(testcase):
        raise helpers.AccessDeniedException()

      if key not in [testcase.fuzzed_keys, testcase.minimized_keys]:
        raise helpers.AccessDeniedException()
    else:
      if not access.has_access():
        raise helpers.AccessDeniedException()

    blob_size = blobs.get_blob_size(key)
    if blob_size > MAX_ALLOWED_CONTENT_SIZE:
      raise helpers.EarlyExitException('Content exceeds max allowed size.', 400)

    # TODO(mbarbella): Workaround for an issue in the Cloud Storage API. Remove
    # once it is fixed properly upstream:
    # https://github.com/googleapis/google-cloud-python/issues/6572
    if blob_size:
      try:
        content = blobs.read_key(key).decode('utf-8', errors='replace')
      except Exception:
        raise helpers.EarlyExitException('Failed to read content.', 400)
    else:
      content = u''

    line_count = len(content.splitlines())
    size = len(content)
    title = '%s, %s' % (utils.get_line_count_string(line_count),
                        utils.get_size_string(size))

    return self.render('viewer.html', {'content': content, 'title': title})



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from clusterfuzz._internal.protos import untrusted_runner_pb2 as clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2


class UntrustedRunnerStub(object):
    """UntrusterRunner service.
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.GetStatus = channel.unary_unary(
                '/UntrustedRunner/GetStatus',
                request_serializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.GetStatusRequest.SerializeToString,
                response_deserializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.GetStatusResponse.FromString,
                )
        self.SetupRegularBuild = channel.unary_unary(
                '/UntrustedRunner/SetupRegularBuild',
                request_serializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.SetupRegularBuildRequest.SerializeToString,
                response_deserializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.SetupBuildResponse.FromString,
                )
        self.RunProcess = channel.unary_unary(
                '/UntrustedRunner/RunProcess',
                request_serializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.RunProcessRequest.SerializeToString,
                response_deserializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.RunProcessResponse.FromString,
                )
        self.RunAndWait = channel.unary_unary(
                '/UntrustedRunner/RunAndWait',
                request_serializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.RunAndWaitRequest.SerializeToString,
                response_deserializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.RunAndWaitResponse.FromString,
                )
        self.CreateDirectory = channel.unary_unary(
                '/UntrustedRunner/CreateDirectory',
                request_serializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.CreateDirectoryRequest.SerializeToString,
                response_deserializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.CreateDirectoryResponse.FromString,
                )
        self.RemoveDirectory = channel.unary_unary(
                '/UntrustedRunner/RemoveDirectory',
                request_serializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.RemoveDirectoryRequest.SerializeToString,
                response_deserializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.RemoveDirectoryResponse.FromString,
                )
        self.ListFiles = channel.unary_unary(
                '/UntrustedRunner/ListFiles',
                request_serializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.ListFilesRequest.SerializeToString,
                response_deserializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.ListFilesResponse.FromString,
                )
        self.CopyFileTo = channel.stream_unary(
                '/UntrustedRunner/CopyFileTo',
                request_serializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.FileChunk.SerializeToString,
                response_deserializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.CopyFileToResponse.FromString,
                )
        self.CopyFileFrom = channel.unary_stream(
                '/UntrustedRunner/CopyFileFrom',
                request_serializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.CopyFileFromRequest.SerializeToString,
                response_deserializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.FileChunk.FromString,
                )
        self.Stat = channel.unary_unary(
                '/UntrustedRunner/Stat',
                request_serializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.StatRequest.SerializeToString,
                response_deserializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.StatResponse.FromString,
                )
        self.UpdateEnvironment = channel.unary_unary(
                '/UntrustedRunner/UpdateEnvironment',
                request_serializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.UpdateEnvironmentRequest.SerializeToString,
                response_deserializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.UpdateEnvironmentResponse.FromString,
                )
        self.ResetEnvironment = channel.unary_unary(
                '/UntrustedRunner/ResetEnvironment',
                request_serializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.ResetEnvironmentRequest.SerializeToString,
                response_deserializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.ResetEnvironmentResponse.FromString,
                )
        self.UpdateSource = channel.unary_unary(
                '/UntrustedRunner/UpdateSource',
                request_serializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.UpdateSourceRequest.SerializeToString,
                response_deserializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.UpdateSourceResponse.FromString,
                )
        self.SymbolizeStacktrace = channel.unary_unary(
                '/UntrustedRunner/SymbolizeStacktrace',
                request_serializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.SymbolizeStacktraceRequest.SerializeToString,
                response_deserializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.SymbolizeStacktraceResponse.FromString,
                )
        self.TerminateStaleApplicationInstances = channel.unary_unary(
                '/UntrustedRunner/TerminateStaleApplicationInstances',
                request_serializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.TerminateStaleApplicationInstancesRequest.SerializeToString,
                response_deserializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.TerminateStaleApplicationInstancesResponse.FromString,
                )
        self.GetFuzzTargets = channel.unary_unary(
                '/UntrustedRunner/GetFuzzTargets',
                request_serializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.GetFuzzTargetsRequest.SerializeToString,
                response_deserializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.GetFuzzTargetsResponse.FromString,
                )
        self.PruneCorpus = channel.unary_unary(
                '/UntrustedRunner/PruneCorpus',
                request_serializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.PruneCorpusRequest.SerializeToString,
                response_deserializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.PruneCorpusResponse.FromString,
                )
        self.ProcessTestcase = channel.unary_unary(
                '/UntrustedRunner/ProcessTestcase',
                request_serializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.ProcessTestcaseRequest.SerializeToString,
                response_deserializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.EngineReproduceResult.FromString,
                )
        self.EngineFuzz = channel.unary_unary(
                '/UntrustedRunner/EngineFuzz',
                request_serializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.EngineFuzzRequest.SerializeToString,
                response_deserializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.EngineFuzzResponse.FromString,
                )
        self.EngineReproduce = channel.unary_unary(
                '/UntrustedRunner/EngineReproduce',
                request_serializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.EngineReproduceRequest.SerializeToString,
                response_deserializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.EngineReproduceResult.FromString,
                )


class UntrustedRunnerServicer(object):
    """UntrusterRunner service.
    """

    def GetStatus(self, request, context):
        """Get information about the worker.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SetupRegularBuild(self, request, context):
        """Set up regular build.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def RunProcess(self, request, context):
        """Run command using process_handler.runProcess
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def RunAndWait(self, request, context):
        """Run command using new_process.ProcessRunner
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CreateDirectory(self, request, context):
        """Create a directory.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def RemoveDirectory(self, request, context):
        """Remove a directory.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ListFiles(self, request, context):
        """List files in a directory.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CopyFileTo(self, request_iterator, context):
        """Copy file from host to worker.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CopyFileFrom(self, request, context):
        """Copy file from worker to host.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Stat(self, request, context):
        """Call stat() on a path.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def UpdateEnvironment(self, request, context):
        """Environment variable changes.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ResetEnvironment(self, request, context):
        """Reset environment variables.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def UpdateSource(self, request, context):
        """Update source.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SymbolizeStacktrace(self, request, context):
        """Symbolize a stacktrace.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def TerminateStaleApplicationInstances(self, request, context):
        """Terminate stale application instances.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetFuzzTargets(self, request, context):
        """libFuzzer/AFL specific: get list of fuzz targets.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def PruneCorpus(self, request, context):
        """libFuzzer specific: corpus pruning
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ProcessTestcase(self, request, context):
        """Engine specific: Do testcase minimization or cleanse.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def EngineFuzz(self, request, context):
        """Engine specific: Do fuzzing.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def EngineReproduce(self, request, context):
        """Engine specific: Do reproduction.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_UntrustedRunnerServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'GetStatus': grpc.unary_unary_rpc_method_handler(
                    servicer.GetStatus,
                    request_deserializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.GetStatusRequest.FromString,
                    response_serializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.GetStatusResponse.SerializeToString,
            ),
            'SetupRegularBuild': grpc.unary_unary_rpc_method_handler(
                    servicer.SetupRegularBuild,
                    request_deserializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.SetupRegularBuildRequest.FromString,
                    response_serializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.SetupBuildResponse.SerializeToString,
            ),
            'RunProcess': grpc.unary_unary_rpc_method_handler(
                    servicer.RunProcess,
                    request_deserializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.RunProcessRequest.FromString,
                    response_serializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.RunProcessResponse.SerializeToString,
            ),
            'RunAndWait': grpc.unary_unary_rpc_method_handler(
                    servicer.RunAndWait,
                    request_deserializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.RunAndWaitRequest.FromString,
                    response_serializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.RunAndWaitResponse.SerializeToString,
            ),
            'CreateDirectory': grpc.unary_unary_rpc_method_handler(
                    servicer.CreateDirectory,
                    request_deserializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.CreateDirectoryRequest.FromString,
                    response_serializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.CreateDirectoryResponse.SerializeToString,
            ),
            'RemoveDirectory': grpc.unary_unary_rpc_method_handler(
                    servicer.RemoveDirectory,
                    request_deserializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.RemoveDirectoryRequest.FromString,
                    response_serializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.RemoveDirectoryResponse.SerializeToString,
            ),
            'ListFiles': grpc.unary_unary_rpc_method_handler(
                    servicer.ListFiles,
                    request_deserializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.ListFilesRequest.FromString,
                    response_serializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.ListFilesResponse.SerializeToString,
            ),
            'CopyFileTo': grpc.stream_unary_rpc_method_handler(
                    servicer.CopyFileTo,
                    request_deserializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.FileChunk.FromString,
                    response_serializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.CopyFileToResponse.SerializeToString,
            ),
            'CopyFileFrom': grpc.unary_stream_rpc_method_handler(
                    servicer.CopyFileFrom,
                    request_deserializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.CopyFileFromRequest.FromString,
                    response_serializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.FileChunk.SerializeToString,
            ),
            'Stat': grpc.unary_unary_rpc_method_handler(
                    servicer.Stat,
                    request_deserializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.StatRequest.FromString,
                    response_serializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.StatResponse.SerializeToString,
            ),
            'UpdateEnvironment': grpc.unary_unary_rpc_method_handler(
                    servicer.UpdateEnvironment,
                    request_deserializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.UpdateEnvironmentRequest.FromString,
                    response_serializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.UpdateEnvironmentResponse.SerializeToString,
            ),
            'ResetEnvironment': grpc.unary_unary_rpc_method_handler(
                    servicer.ResetEnvironment,
                    request_deserializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.ResetEnvironmentRequest.FromString,
                    response_serializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.ResetEnvironmentResponse.SerializeToString,
            ),
            'UpdateSource': grpc.unary_unary_rpc_method_handler(
                    servicer.UpdateSource,
                    request_deserializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.UpdateSourceRequest.FromString,
                    response_serializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.UpdateSourceResponse.SerializeToString,
            ),
            'SymbolizeStacktrace': grpc.unary_unary_rpc_method_handler(
                    servicer.SymbolizeStacktrace,
                    request_deserializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.SymbolizeStacktraceRequest.FromString,
                    response_serializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.SymbolizeStacktraceResponse.SerializeToString,
            ),
            'TerminateStaleApplicationInstances': grpc.unary_unary_rpc_method_handler(
                    servicer.TerminateStaleApplicationInstances,
                    request_deserializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.TerminateStaleApplicationInstancesRequest.FromString,
                    response_serializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.TerminateStaleApplicationInstancesResponse.SerializeToString,
            ),
            'GetFuzzTargets': grpc.unary_unary_rpc_method_handler(
                    servicer.GetFuzzTargets,
                    request_deserializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.GetFuzzTargetsRequest.FromString,
                    response_serializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.GetFuzzTargetsResponse.SerializeToString,
            ),
            'PruneCorpus': grpc.unary_unary_rpc_method_handler(
                    servicer.PruneCorpus,
                    request_deserializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.PruneCorpusRequest.FromString,
                    response_serializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.PruneCorpusResponse.SerializeToString,
            ),
            'ProcessTestcase': grpc.unary_unary_rpc_method_handler(
                    servicer.ProcessTestcase,
                    request_deserializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.ProcessTestcaseRequest.FromString,
                    response_serializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.EngineReproduceResult.SerializeToString,
            ),
            'EngineFuzz': grpc.unary_unary_rpc_method_handler(
                    servicer.EngineFuzz,
                    request_deserializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.EngineFuzzRequest.FromString,
                    response_serializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.EngineFuzzResponse.SerializeToString,
            ),
            'EngineReproduce': grpc.unary_unary_rpc_method_handler(
                    servicer.EngineReproduce,
                    request_deserializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.EngineReproduceRequest.FromString,
                    response_serializer=clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.EngineReproduceResult.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'UntrustedRunner', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class UntrustedRunner(object):
    """UntrusterRunner service.
    """

    @staticmethod
    def GetStatus(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/UntrustedRunner/GetStatus',
            clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.GetStatusRequest.SerializeToString,
            clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.GetStatusResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def SetupRegularBuild(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/UntrustedRunner/SetupRegularBuild',
            clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.SetupRegularBuildRequest.SerializeToString,
            clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.SetupBuildResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def RunProcess(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/UntrustedRunner/RunProcess',
            clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.RunProcessRequest.SerializeToString,
            clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.RunProcessResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def RunAndWait(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/UntrustedRunner/RunAndWait',
            clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.RunAndWaitRequest.SerializeToString,
            clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.RunAndWaitResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def CreateDirectory(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/UntrustedRunner/CreateDirectory',
            clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.CreateDirectoryRequest.SerializeToString,
            clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.CreateDirectoryResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def RemoveDirectory(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/UntrustedRunner/RemoveDirectory',
            clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.RemoveDirectoryRequest.SerializeToString,
            clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.RemoveDirectoryResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ListFiles(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/UntrustedRunner/ListFiles',
            clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.ListFilesRequest.SerializeToString,
            clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.ListFilesResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def CopyFileTo(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_unary(request_iterator, target, '/UntrustedRunner/CopyFileTo',
            clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.FileChunk.SerializeToString,
            clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.CopyFileToResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def CopyFileFrom(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/UntrustedRunner/CopyFileFrom',
            clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.CopyFileFromRequest.SerializeToString,
            clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.FileChunk.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Stat(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/UntrustedRunner/Stat',
            clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.StatRequest.SerializeToString,
            clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.StatResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def UpdateEnvironment(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/UntrustedRunner/UpdateEnvironment',
            clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.UpdateEnvironmentRequest.SerializeToString,
            clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.UpdateEnvironmentResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ResetEnvironment(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/UntrustedRunner/ResetEnvironment',
            clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.ResetEnvironmentRequest.SerializeToString,
            clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.ResetEnvironmentResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def UpdateSource(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/UntrustedRunner/UpdateSource',
            clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.UpdateSourceRequest.SerializeToString,
            clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.UpdateSourceResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def SymbolizeStacktrace(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/UntrustedRunner/SymbolizeStacktrace',
            clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.SymbolizeStacktraceRequest.SerializeToString,
            clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.SymbolizeStacktraceResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def TerminateStaleApplicationInstances(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/UntrustedRunner/TerminateStaleApplicationInstances',
            clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.TerminateStaleApplicationInstancesRequest.SerializeToString,
            clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.TerminateStaleApplicationInstancesResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetFuzzTargets(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/UntrustedRunner/GetFuzzTargets',
            clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.GetFuzzTargetsRequest.SerializeToString,
            clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.GetFuzzTargetsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def PruneCorpus(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/UntrustedRunner/PruneCorpus',
            clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.PruneCorpusRequest.SerializeToString,
            clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.PruneCorpusResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ProcessTestcase(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/UntrustedRunner/ProcessTestcase',
            clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.ProcessTestcaseRequest.SerializeToString,
            clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.EngineReproduceResult.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def EngineFuzz(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/UntrustedRunner/EngineFuzz',
            clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.EngineFuzzRequest.SerializeToString,
            clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.EngineFuzzResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def EngineReproduce(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/UntrustedRunner/EngineReproduce',
            clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.EngineReproduceRequest.SerializeToString,
            clusterfuzz_dot___internal_dot_protos_dot_untrusted__runner__pb2.EngineReproduceResult.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""py_unittest.py runs tests under src/appengine and butler/tests"""
import time
import urllib.error
import urllib.request

from clusterfuzz._internal.tests.test_libs import test_utils
from local.butler import common
from local.butler import constants

RUN_SERVER_TIMEOUT = 120


def execute(_):
  """Run integration tests."""
  command = 'run_server'
  indicator = b'Booting worker'

  try:
    lines = []
    server = common.execute_async(
        'python -u butler.py {} --skip-install-deps'.format(command))
    test_utils.wait_for_emulator_ready(
        server,
        command,
        indicator,
        timeout=RUN_SERVER_TIMEOUT,
        output_lines=lines)

    # Sleep a small amount of time to ensure the server is definitely ready.
    time.sleep(1)

    # Call setup ourselves instead of passing --bootstrap since we have no idea
    # when that finishes.
    # TODO(ochang): Make bootstrap a separate butler command and just call that.
    common.execute(
        ('python butler.py run setup '
         '--non-dry-run --local --config-dir={config_dir}'
        ).format(config_dir=constants.TEST_CONFIG_DIR),
        exit_on_error=False)

    request = urllib.request.urlopen('http://' + constants.DEV_APPSERVER_HOST)
    request.read()  # Raises exception on error
  except Exception:
    print('Error occurred:')
    print(b''.join(lines))
    raise
  finally:
    server.terminate()

  # TODO(ochang): Test that bot runs, and do a basic fuzzing session to ensure
  # things work end to end.
  print('All end-to-end integration tests passed.')



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""filters tests"""

import unittest

import mock

from libs import filters
from libs import helpers


class ExtractKeywordFieldTest(unittest.TestCase):
  """Test extact_keyword_field."""

  def _test(self, keyword, expected_rest, expected_value):
    rest, value = filters.extract_keyword_field(keyword, 'field')
    self.assertEqual(expected_rest, rest)
    self.assertEqual(expected_value, value)

  def test_empty(self):
    """Test empty."""
    self._test('', '', None)

  def test_no_field(self):
    """There is no keyword field."""
    self._test('key1 key2', 'key1 key2', None)

  def test_extract_field(self):
    """There is a field."""
    self._test('keyfield:value field:value key', 'keyfield:value key', 'value')
    self._test('keyfield:value field: key', 'keyfield:value key', '')
    self._test('keyfield:value field:value" key', 'keyfield:value key',
               'value"')
    self._test("keyfield:value field:'value key", 'keyfield:value key',
               "'value")

  def test_extract_field_single_quote(self):
    """There is a field with single quotes."""
    self._test("keyfield:value field:'val ue' key", 'keyfield:value key',
               'val ue')
    self._test("keyfield:value field:'' key", 'keyfield:value key', '')
    self._test("keyfield:value field:' \"a\" ' key", 'keyfield:value key',
               ' "a" ')

  def test_extract_field_double_quote(self):
    """There is a field with double quotes."""
    self._test('keyfield:value field:"val ue" key', 'keyfield:value key',
               'val ue')
    self._test('keyfield:value field:"" key', 'keyfield:value key', '')
    self._test('keyfield:value field:" \'a\' " key', 'keyfield:value key',
               " \'a\' ")


class SimpleFilterTest(unittest.TestCase):
  """Test SimpleFilter."""

  def setUp(self):
    self.query = mock.MagicMock()

  def test_empty(self):
    """Test empty value."""
    fltr = filters.SimpleFilter(
        'field', 'param', transformers=[lambda v: '1'], required=False)
    fltr.add(self.query, {'param': ''})
    self.query.filter.assert_not_called()

  def test_empty_required(self):
    """Test required value."""
    fltr = filters.SimpleFilter(
        'field', 'param', transformers=[lambda v: '1'], required=True)

    with self.assertRaises(helpers.EarlyExitException) as cm:
      fltr.add(self.query, {'param': ''})

    self.assertEqual("'param' is required.", str(cm.exception))
    self.assertEqual(400, cm.exception.status)
    self.query.filter.assert_not_called()

  def test_transform(self):
    """Test transform."""
    fltr = filters.SimpleFilter(
        'field', 'param', transformers=[lambda v: '1'], required=False)
    fltr.add(self.query, {'param': 'test'})
    self.query.filter.assert_called_once_with('field', '1')


class StringTest(unittest.TestCase):
  """Test String."""

  def setUp(self):
    self.filter = filters.String('field', 'param')
    self.query = mock.Mock()

  def test_empty(self):
    """Test empty."""
    self.filter.add(self.query, {})
    self.query.assert_not_called()

  def test_get(self):
    """Test get stripped string."""
    self.filter.add(self.query, {'param': ' aAa '})
    self.query.assert_has_calls([mock.call.filter('field', 'aAa')])


class KeywordTest(unittest.TestCase):
  """Test Keyword."""

  def setUp(self):
    fltrs = [
        filters.String('string_field', 'string_param'),
        filters.Int('int_field', 'int_param')
    ]
    self.filter = filters.Keyword(fltrs, 'field', 'param')
    self.query = mock.Mock()

  def test_get(self):
    """Test get keyword."""
    self.filter.add(self.query,
                    {'param': 'aaa bbb string_param:val int_param:234'})
    self.query.assert_has_calls([
        mock.call.filter('string_field', 'val'),
        mock.call.filter('int_field', 234),
        mock.call.filter('field', 'aaa'),
        mock.call.filter('field', 'bbb')
    ])


class NegativeBooleanTest(unittest.TestCase):
  """Test Boolean."""

  def setUp(self):
    self.filter = filters.NegativeBoolean('field', 'param')
    self.query = mock.Mock()

  def test_yes(self):
    """Test yes."""
    self.filter.add(self.query, {'param': 'yes'})
    self.query.assert_has_calls([mock.call.filter('field', False)])

  def test_no(self):
    """Test no."""
    self.filter.add(self.query, {'param': 'no'})
    self.query.assert_has_calls([mock.call.filter('field', True)])


class BooleanTest(unittest.TestCase):
  """Test Boolean."""

  def setUp(self):
    self.filter = filters.Boolean('field', 'param')
    self.query = mock.Mock()

  def test_empty(self):
    """Test empty."""
    self.filter.add(self.query, {'param': ''})
    self.query.assert_not_called()

  def test_yes(self):
    """Test yes."""
    self.filter.add(self.query, {'param': 'yes'})
    self.query.assert_has_calls([mock.call.filter('field', True)])

  def test_no(self):
    """Test no."""
    self.filter.add(self.query, {'param': 'no'})
    self.query.assert_has_calls([mock.call.filter('field', False)])

  def test_exception(self):
    """Test exception."""
    with self.assertRaises(helpers.EarlyExitException):
      self.filter.add(self.query, {'param': 'wsdljf'})
    self.query.assert_not_called()


class IntTest(unittest.TestCase):
  """Test String."""

  def setUp(self):
    self.filter = filters.Int('field', 'param')
    self.query = mock.Mock()

  def test_empty(self):
    """Test empty."""
    self.filter.add(self.query, {'param': ''})
    self.query.assert_not_called()

  def test_get(self):
    """Test get int."""
    self.filter.add(self.query, {'param': '0'})
    self.query.assert_has_calls([mock.call.filter('field', 0)])

  def test_exception(self):
    """Test exception."""
    with self.assertRaises(helpers.EarlyExitException):
      self.filter.add(self.query, {'param': 'wsdljf'})
    self.query.assert_not_called()

  def test_operator(self):
    fltr = filters.Int('field', 'param', operator='>')
    fltr.add(self.query, {'param': '0'})
    self.query.assert_has_calls([mock.call.filter('field', 0, operator='>')])


class AddTest(unittest.TestCase):
  """Test add."""

  def setUp(self):
    self.params = {}
    self.query = mock.Mock()
    self.filters = [
        filters.String('string_field', 'string_param'),
        filters.Int('int_field', 'int_param'),
        filters.Boolean('bool_field', 'bool_param'),
    ]

  def test_no_field(self):
    """Test no field."""
    filters.add(self.query, self.params, self.filters)
    self.query.filter.assert_has_calls([])

  def test_multiple_fields(self):
    """Test add multiple filters."""
    self.params['string_param'] = 'value'
    self.params['int_param'] = '123'
    self.params['bool_param'] = 'yes'

    filters.add(self.query, self.params, self.filters)

    self.query.assert_has_calls([
        mock.call.filter('string_field', 'value'),
        mock.call.filter('int_field', 123),
        mock.call.filter('bool_field', True),
    ])
